# Deformation energy module
#
# Written by Konrad Hinsen
# last revision: 1999-3-19
#

from MMTK_forcefield import NonbondedList
from MMTK_deformation import deformation, reduceDeformation, \
                             reduceFiniteDeformation
import Variables
import Numeric

#
# Deformation energy evaluations
#
class DeformationEvaluationFunction:

    def __init__(self, universe, fc_length = 0.7, cutoff = 1.2,
		 factor = 46401.746122):
	self.universe = universe
	self.fc_length = fc_length
	self.cutoff = cutoff
	self.factor = factor

	nothing = Numeric.zeros((0,2), Numeric.Int)
	self.pairs = NonbondedList(nothing, nothing, nothing,
				   universe._spec, cutoff)
	self.pairs.update(universe.configuration().array)
	self.normalize = 0

    def newConfiguration(self):
	self.pairs.update(self.universe.configuration().array)


class DeformationFunction(DeformationEvaluationFunction):

    def __call__(self, vector):
	conf = self.universe.configuration()
	r = Variables.ParticleScalar(self.universe)
	l = deformation(conf.array, vector.array, self.pairs,
			None, r.array, self.fc_length, self.factor,
			self.normalize)
	return r

class NormalizedDeformationFunction(DeformationFunction):

    def __init__(self, *args):
	apply(DeformationFunction.__init__, (self, ) + args)
	self.normalize = 1

class FiniteDeformationFunction(DeformationEvaluationFunction):

    def __call__(self, vector):
	conf = self.universe.configuration()
	vector = vector-conf
	r = Variables.ParticleScalar(self.universe)
	l = deformation(conf.array, vector.array, self.pairs,
			None, r.array, self.fc_length, self.factor, 0, 1)
	return r

class DeformationEnergyFunction(DeformationEvaluationFunction):

    def __call__(self, vector, gradients = None):
	conf = self.universe.configuration()
	g = None
	if gradients is not None:
	    if Variables.isUniverseVariable(gradients):
		g = gradients
	    elif type(gradients) == Numeric.arraytype:
		g = Variables.ParticleVector(self.universe, gradients)
	    elif gradients:
		g = Variables.ParticleVector(self.universe)
	if g is None:
	    g_array = None
	else:
	    g_array = g.array
	l = deformation(conf.array, vector.array, self.pairs,
			g_array, None, self.fc_length, self.factor,
			self.normalize)
	if g is None:
	    return l
	else:
	    return l, g

class NormalizedDeformationEnergyFunction(DeformationEnergyFunction):

    def __init__(self, *args):
	apply(DeformationEnergyFunction.__init__, (self,) + args)
	self.normalize = 1

class FiniteDeformationEnergyFunction(DeformationEvaluationFunction):

    def __call__(self, vector, gradients = None):
	conf = self.universe.configuration()
	g = None
	if gradients is not None:
	    if Variables.isUniverseVariable(gradients):
		g = gradients
	    elif type(gradients) == Numeric.arraytype:
		g = Variables.ParticleVector(self.universe, gradients)
	    elif gradients:
		g = Variables.ParticleVector(self.universe)
	if g is None:
	    g_array = None
	else:
	    g_array = g.array
	l = deformation(conf.array, vector.array, self.pairs,
			g_array, None, self.fc_length, self.factor,
			0, 1)
	if g is None:
	    return l
	else:
	    return l, g

#
# Deformation energy minimization
#
class DeformationReducer(DeformationEvaluationFunction):

    def __call__(self, vector, niter):
	conf = self.universe.configuration()
	reduceDeformation(conf.array, vector.array, self.pairs,
			  self.fc_length, self.factor, niter)

class FiniteDeformationReducer(DeformationEvaluationFunction):

    def __call__(self, vector, rms_reduction):
	conf = self.universe.configuration()
	vector = vector-conf
	reduceFiniteDeformation(conf.array, vector.array, self.pairs,
				self.fc_length, self.factor, rms_reduction)
	return conf+vector
