# This module implements MD integrators
#
# Written by Konrad Hinsen
# last revision: 1999-3-19
#

import Environment, Features, Trajectory, Units
import MMTK_dynamics
import MMTK_forcefield
import Numeric, operator

#
# Integrator base class
#
class Integrator(Trajectory.TrajectoryGenerator):

    def __init__(self, universe, options):
	Trajectory.TrajectoryGenerator.__init__(self, universe, options)

    default_options = {'steps': 100, 'delta_t': 1.*Units.fs,
		       'options': []}

    available_data = ['configuration', 'velocities', 'gradients',
		      'energy', 'thermodynamic', 'time', 'auxiliary']

    restart_data = ['configuration', 'velocities', 'energy',
                    'thermodynamic', 'auxiliary']

    def __call__(self, options):
	raise AttributeError

#
# Velocity-Verlet integrator
#
class VelocityVerletIntegrator(Integrator):

    def __init__(self, universe, **options):
	Integrator.__init__(self, universe, options)
	self.features = [Features.FixedParticleFeature,
			 Features.NoseThermostatFeature,
			 Features.AndersenBarostatFeature,
                         Features.DistanceConstraintsFeature]

    def __call__(self, **options):
	self.setCallOptions(options)
	used_features = Features.checkFeatures(self, self.universe)
	configuration = self.universe.configuration()
	velocities = self.universe.velocities()
	if velocities is None:
	    raise ValueError, "no velocities"
	masses = self.universe.masses()
	fixed = self.universe.getAtomBooleanArray('fixed')
	evaluator = self.universe.energyEvaluator().evaluator.CEvaluator()
	evaluator = MMTK_forcefield.TotalEvaluator(evaluator.allTerms())
        constraints, const_distances_sq, c_blocks = \
                     _constraintArrays(self.universe)
                                                  
        type = 'NVE'
	if Features.NoseThermostatFeature in used_features:
            type = 'NVT'
	    for object in self.universe._environment:
		if object.__class__ is Environment.NoseThermostat:
		    thermostat = object
		    break
            t_parameters = thermostat.parameters
            t_coordinates = thermostat.coordinates
        else:
            t_parameters = Numeric.zeros((0,), Numeric.Float)
            t_coordinates = Numeric.zeros((2,), Numeric.Float)
        if Features.AndersenBarostatFeature in used_features:
            if self.universe.cellVolume() is None:
                raise ValueError, "Barostat requires finite volume universe"
            if type == 'NVE':
                type = 'NPH'
            else:
                type = 'NPT'
	    for object in self.universe._environment:
		if object.__class__ is Environment.AndersenBarostat:
		    barostat = object
		    break
            b_parameters = barostat.parameters
            b_coordinates = barostat.coordinates
        else:
            b_parameters = Numeric.zeros((0,), Numeric.Float)
            b_coordinates = Numeric.zeros((1,), Numeric.Float)

        MMTK_dynamics.integrateVV(self.universe,
                                  configuration.array, velocities.array,
                                  masses.array, fixed.array, evaluator,
                                  constraints, const_distances_sq, c_blocks,
                                  t_parameters, t_coordinates,
                                  b_parameters, b_coordinates,
                                  self.getOption('delta_t'),
                                  self.getOption('steps'),
                                  self.getActions(),
                                  type + ' dynamics trajectory with ' +
                                  self.optionString(['delta_t', 'steps']))

#
# Velocity scaling, removal of global translation/rotation
#
class VelocityScaler(Trajectory.TrajectoryAction):

    def __init__(self, temperature, window=0., first=0, last=None, skip=1):
        Trajectory.TrajectoryAction.__init__(self, first, last, skip)
	self.parameters = Numeric.array([temperature, window], Numeric.Float)
	self.Cfunction = MMTK_dynamics.scaleVelocities

class Heater(Trajectory.TrajectoryAction):

    def __init__(self, temp1, temp2, gradient, first=0, last=None, skip=1):
        Trajectory.TrajectoryAction.__init__(self, first, last, skip)
	self.parameters = Numeric.array([temp1, temp2, gradient],
                                        Numeric.Float)
	self.Cfunction = MMTK_dynamics.heat

class BarostatReset(Trajectory.TrajectoryAction):

    def __init__(self, first=0, last=None, skip=1):
        Trajectory.TrajectoryAction.__init__(self, first, last, skip)
        self.parameters = Numeric.zeros((0,), Numeric.Float)
        self.Cfunction = MMTK_dynamics.resetBarostat

class TranslationRemover(Trajectory.TrajectoryAction):

    def __init__(self, first=0, last=None, skip=1):
        Trajectory.TrajectoryAction.__init__(self, first, last, skip)
	self.parameters = None
	self.Cfunction = MMTK_dynamics.removeTranslation

class RotationRemover(Trajectory.TrajectoryAction):

    def __init__(self, first=0, last=None, skip=1):
        Trajectory.TrajectoryAction.__init__(self, first, last, skip)
	self.parameters = None
	self.Cfunction = MMTK_dynamics.removeRotation

#
# Construct constraint arrays
#
def _constraintArrays(universe):
    nc = universe.numberOfDistanceConstraints()
    constraints = Numeric.zeros((nc, 2), Numeric.Int)
    const_distances_sq = Numeric.zeros((nc, ), Numeric.Float)
    if nc > 0:
        i = 0
        c_blocks = [0]
        for o in universe.objectList():
            for c in o.distanceConstraintList():
                constraints[i, 0] = c[0].index
                constraints[i, 1] = c[1].index
                const_distances_sq[i] = c[2]*c[2]
                i = i + 1
            c_blocks.append(i)
        c_blocks = Numeric.array(c_blocks)
    else:
        c_blocks = Numeric.zeros((1,), Numeric.Int)
    return constraints, const_distances_sq, c_blocks

#
# Enforce distance constraints for current configuration
#
def enforceConstraints(universe, configuration=None):
    constraints, const_distances_sq, c_blocks = _constraintArrays(universe)
    if len(constraints) == 0:
        return
    if configuration is None:
        configuration = universe.configuration()
    MMTK_dynamics.enforceConstraints(universe._spec, configuration.array,
                                     universe.masses().array,
                                     constraints, const_distances_sq, c_blocks)

#
# Project velocity vector onto the constraint surface
#
def projectVelocities(universe, velocities):
    constraints, const_distances_sq, c_blocks = _constraintArrays(universe)
    if len(constraints) == 0:
        return
    MMTK_dynamics.projectVelocities(universe._spec,
                                    universe.configuration().array,
                                    velocities.array, universe.masses().array,
                                    constraints, const_distances_sq, c_blocks)
