# This file provides a force field, using data from TinkerParamData.
#
# Written by Peter McCluskey
#

from MMTK.ForceFields.BondedInteractions import BondedForceField, BondedFFEvaluator
from MMTK.ForceFields.NonBondedInteractions import NonBondedForceField
from MMTK.ForceFields.ForceField import CompoundForceField
from TinkerParamData import TinkerParamData
from MMTK.CompositeObject import isCompositeChemicalObject
from Vector import Vector
import copy
from MMTK.Units import deg
from MMTK.Bonds import BondAngle
from MMTK import Utility
import string

#
# Common base class for generic force field terms
#
class GenericForceFieldTerm:

    def __init__(self, filename, file_format = 'tinker'):
	self.short_name = filename
	if file_format == 'tinker':
	    self.dataset = TinkerParamData(filename)
	else:
	    raise RuntimeError, 'GenericForceFieldTerm - ' + file_format + ' not implemented'

    def deduceAtomType(self, a, blist = None, num_dangling = None):
	if blist is None:
	    blist = a.bondedTo()
	if num_dangling is None:
	    num_dangling = a.numDanglingBondsTo()
	key = (string.upper(a.symbol), len(blist) + num_dangling)
	if self.dataset.atom_type_map.has_key(key):
	    type_list = self.dataset.atom_type_map[key]
	    if len(type_list) == 1:
		return type_list[0]
	    return self.dataset.deduceAtomType(a, blist, num_dangling)
	print 'num_dangling',num_dangling,hex(id(a)),blist,a.fullName()
	msg = 'Unable to deduce type for atom ' + `a` +\
	      ' with ' + `len(blist) + num_dangling` + ' bonds'
	raise KeyError, msg

    def collectAtomTypesAndIndices(self, universe, global_data):
	if not hasattr(global_data, self.dataset.atom_type_property):
	    type = {}
	    for o in universe:
		for a in o.atomList():
		    try:
			type[a] = o.getAtomProperty(a, self.dataset.atom_type_property)
		    except (KeyError, TypeError):
			type[a] = self.deduceAtomType(a)
	    global_data.atom_type = type
	    setattr(global_data, self.dataset.atom_type_property, type)

    def deduceAtomCharge(self, a, object):
	try:
	    return object.getAtomProperty(a, 'amber_charge')
	except (KeyError, TypeError, AttributeError):
	    return 0			# lousy hack
	#raise KeyError, 'Unable to deduce charge for atom ' + `a`

    def collectCharges(self, universe, global_data):
	if not hasattr(global_data, self.dataset.charge_property):
	    charge = {}
	    for o in universe:
		for a in o.atomList():
		    try:
			charge[a] = o.getAtomProperty(a, self.dataset.charge_property)
		    except (KeyError, TypeError, AttributeError):
			charge[a] = self.deduceAtomCharge(a, o)
	    global_data.charge = charge
	    setattr(global_data, self.dataset.charge_property, type)

    def _atomType(self, object, atom, global_data):
	return global_data.atom_type[atom]

    def _charge(self, object, atom, global_data):
	return global_data.charge[atom]

    def getOrSetAtomType(self, atom, object):
	if atom is None:
	    return None
	try:
	    try:
		type1 = object.getAtomProperty(atom, self.dataset.atom_type_property)
		#print 'getOrSetAtomType get',hex(id(atom)),type1,hex(id(object)),atom.fullName()
		return type1
	    except AttributeError:
		if not hasattr(object, self.dataset.atom_type_property):
		    setattr(object, self.dataset.atom_type_property, {})
		raise TypeError
	except TypeError:
	    type1 = self.deduceAtomType(atom)
	    object.setAtomProperty(atom, self.dataset.atom_type_property, type1)
	    #print 'getOrSetAtomType set',hex(id(atom)),type1,hex(id(object))
	    return type1

    def defaultBondLength(self, bond, object):
	type1 = self.getOrSetAtomType(bond.a1, object)
	type2 = self.getOrSetAtomType(bond.a2, object)
	try:
	    (l, k) = self.dataset.bondParameters(type1, type2)
	except KeyError:
	    print 'defaultBondLength types', type1, type2,self.dataset.atom_type_property,hex(id(bond.a1)),hex(id(bond.a2))
	    msg = self.dataset.short_filename + ' has no bond length data for '\
		  + `bond` + ', types ' + `type1` + ',' + `type2`
	    raise TypeError, msg
	return l

    def defaultBondAngle(self, bond1, bond2, object):
	type1 = self.getOrSetAtomType(bond1.a1, object)
	type2 = self.getOrSetAtomType(bond1.a2, object)
	if not bond2.hasAtom(bond1.a2):
	    (type1, type2) = (type2, type1)
	if bond1.hasAtom(bond2.a1):
	    third_atom = bond2.a2
	else:
	    third_atom = bond2.a1
	if third_atom is None:
	    (a, e) = self.dataset.bondAngleAvgParameters(type1, type2)
	    return a
	type3 = self.getOrSetAtomType(third_atom, object)
	ba = BondAngle(bond1, bond2)
	num_h = self.countOtherHydrogens(ba, object)
	try:
	    try:
		(a, e) = self.dataset.bondAngleParameters(type1, type2, type3, num_h)
		#print bond1.a1.name,bond1.a2.name,third_atom.name
		#print 'defaultBondAngle %.4f %.4f %d %d %d' % (a/deg, e, type1, type2, type3)
	    except IndexError:
		ca = ba.ca
		print 'types',type1, type2, type3
		Utility.warning('Atom ' + `ca` + ' has more hydrogens(' + `num_h`\
				+ ') than forcefield expects outside of BondAngle '\
				+ `ba`)
		(a, e) = self.dataset.bondAngleParameters(type1, type2, type3, 0)
	except KeyError:
		a = 109.5*deg
	return a

		# Count the number of hydrogens attached to an angle's central
		# atom but not in either of the bonds forming the angle
    def countOtherHydrogens(self, angle, object):
         num_h = 0
	 for a in angle.ca.bondedTo():
	    if not (a is angle.a1 or a is angle.a2):
		ia = self.getOrSetAtomType(a, object)
		if self.dataset.atomic_number[ia] == 1:
		    num_h = num_h + 1
	 return num_h

    def bondOrder(self, atom, object):
	type1 = self.getOrSetAtomType(atom, object)
	return self.dataset.bondOrder(type1)
	

#
# Bonded interactions
#
class GenericBondedForceField(GenericForceFieldTerm, BondedForceField):

    def __init__(self, filename):
	GenericForceFieldTerm.__init__(self, filename)
	BondedForceField.__init__(self, filename)

    def evaluator(self, universe, subset1, subset2, global_data):
	self.collectAtomTypesAndIndices(universe, global_data)
	return BondedForceField.evaluator(self, universe, subset1, subset2,
					  global_data)

    def addBondTerm(self, data, bond, object, global_data):
	a1 = bond.a1
	a2 = bond.a2
	i1 = a1.index
	i2 = a2.index
	global_data.add('excluded_pairs', (i1, i2))
	t1 = global_data.atom_type[a1]
	t2 = global_data.atom_type[a2]
	try:
	    p = self.dataset.bondParameters(t1, t2)
	except KeyError:
	    raise KeyError, 'No parameters for bond ' + `a1` +  '--' + `a2`
	if p is not None:
	    data.add('bonds', (i1, i2) + p)

    def addBondAngleTerm(self, data, angle, object, global_data):
	a1 = angle.a1
	a2 = angle.a2
	ca = angle.ca
	i1 = a1.index
	i2 = a2.index
	ic = ca.index
	global_data.add('excluded_pairs', (i1, i2))
	t1 = global_data.atom_type[a1]
	t2 = global_data.atom_type[a2]
	tc = global_data.atom_type[ca]
	num_h = self.countOtherHydrogens(angle, object)
	try:
	    try:
		p = self.dataset.bondAngleParameters(t1, tc, t2, num_h)
	    except IndexError:
		Utility.warning('Atom ' + `ca` + ' has more hydrogens(' + `num_h`\
				+ ') than forcefield expects outside of BondAngle '\
				+ `angle`)
		p = self.dataset.bondAngleParameters(t1, tc, t2, 0)
	except KeyError:
	    raise KeyError, 'No parameters for angle ' + `a1` +  \
		  '--' + `ca` + '--' + `a2`
	if p is not None:
	    data.add('angles', (i1, ic, i2) + p)

    def addDihedralTerm(self, data, dihedral, object, global_data):
	a1 = dihedral.a1
	a2 = dihedral.a2
	a3 = dihedral.a3
	a4 = dihedral.a4
	i1 = a1.index
	i2 = a2.index
	i3 = a3.index
	i4 = a4.index
	global_data.add('1_4_pairs', (i1, i4))
	t1 = global_data.atom_type[a1]
	t2 = global_data.atom_type[a2]
	t3 = global_data.atom_type[a3]
	t4 = global_data.atom_type[a4]
	terms = self.dataset.dihedralParameters(t1, t2, t3, t4)
	if terms is not None:
	    for p in terms:
		if p[2] != 0.:
		    data.add('dihedrals', (i1, i2, i3, i4) + p)

    def addImproperTerm(self, data, improper, object, global_data):
	a1 = improper.a1
	a2 = improper.a2
	a3 = improper.a3
	a4 = improper.a4
	i1 = a1.index
	i2 = a2.index
	i3 = a3.index
	i4 = a4.index
	t1 = global_data.atom_type[a1]
	t2 = global_data.atom_type[a2]
	t3 = global_data.atom_type[a3]
	t4 = global_data.atom_type[a4]
	terms = self.dataset.improperParameters(t1, t2, t3, t4)
	if terms is not None:
	    atoms = [(t2,i2), (t3,i3), (t4,i4)]
	    atoms.sort(_order)
	    i2, i3, i4 = tuple(map(lambda t: t[1], atoms))
	    for p in terms:
		if p[2] != 0.:
		    data.add('dihedrals', (i2, i3, i1, i4) + p)

    def evalObject(self, global_data):
	return BondedFFEvaluator(self.data, global_data)

    def bonds(self, global_data):
	return self.data.get('bonds')
    def angles(self, global_data):
	return self.data.get('angles')
    def dihedrals(self, global_data):
	return self.data.get('dihedrals')

def _order(a1, a2):
    ret = cmp(a1[0], a2[0])
    if ret == 0:
	ret = cmp(a1[1], a2[1])
    return ret


#
# Nonbonded interactions
#

class GenericNonbondedForceField(GenericForceFieldTerm, NonBondedForceField):

    def __init__(self, filename):
	GenericForceFieldTerm.__init__(self, filename)
	self.arguments = ()
            
    def charge(self, atoms):
	total = 0.
	for a in atoms.atomList():
	    charge = a.topLevelChemicalObject() \
		     .getAtomProperty(a, self.dataset.charge_property)
	    total = total + charge
	return total

    def dipole(self, atoms, reference = None):
	if reference is None:
	    reference = atoms.centerOfMass()
	total = Vector(0., 0., 0.)
	for a in atoms.atomList():
	    charge = a.topLevelChemicalObject() \
		     .getAtomProperty(a, self.dataset.charge_property)
	    total = total + charge*a.position()-reference
	return total

    def evaluator(self, universe, subset1, subset2, global_data):
	self.collectAtomTypesAndIndices(universe, global_data)
	self.collectCharges(universe, global_data)
	return CompoundForceField().evaluator(universe, subset1,
					      subset2, global_data)

#
# The total force field
#
class GenericForceField(GenericForceFieldTerm, CompoundForceField):

    def __init__(self, filename):
	GenericForceFieldTerm.__init__(self, filename)
	self.arguments = (filename, )
	self.bonded = GenericBondedForceField(filename)
        self.nonbonded = apply(GenericNonbondedForceField, self.arguments)
        apply(CompoundForceField.__init__,
              (self, self.bonded, self.nonbonded))

    def evaluator(self, universe, subset1, subset2, global_data):
	self.collectAtomTypesAndIndices(universe, global_data)
	return CompoundForceField.evaluator(self, universe, subset1, subset2,
					    global_data)

    def charge(self, atoms):
	return self.nonbonded.charge(atoms)

    def dipole(self, atoms, reference = None):
	return self.nonbonded.dipole(atoms, reference)

