# This module implements subspaces for motion analysis etc.
#
# Written by Konrad Hinsen
# last revision: 1999-2-23
#

import Utility, Variables
from Scientific.Geometry import Vector
import Numeric


class ParticleVectorSet:

    def __init__(self, universe, data):
        self.universe = universe
        if type(data) == Numeric.arraytype:
            self.n = data.shape[0]
            self.array = data
        else:
            self.n = data
            self.array = Numeric.zeros((self.n, universe.numberOfAtoms(), 3),
                                       Numeric.Float)

    def len(self):
        return self.n

    def __getitem__(self, i):
        if i >= self.n:
            raise IndexError
        return Variables.ParticleVector(universe, self.array[i])


class Subspace:

    def __init__(self, universe, vectors):
        self.universe = universe
        self.vectors = vectors
        self._basis = None

    def __add__(self, other):
        return Subspace(self.vectors+other.vectors)

    def basis(self):
        if self._basis is None:
            try:
                from lapack_dge import dgesvd
            except ImportError:
                from lapack_lite import dgesvd
            basis = Numeric.array(self.vectors, Numeric.Float)
            shape = basis.shape
            nvectors = shape[0]
            natoms = shape[1]
            basis.shape = (nvectors, 3*natoms)
            sv = Numeric.zeros((min(nvectors, 3*natoms),), Numeric.Float)
            work = Numeric.zeros((max(3*min(3*natoms, nvectors)
                                            + max(3*natoms, nvectors),
                                      5*min(3*natoms, nvectors)-4),),
                                 Numeric.Float)
            dummy = Numeric.zeros((1,), Numeric.Float)
            result = dgesvd('O', 'N', 3*natoms, nvectors, basis, 3*natoms,
                            sv, dummy, 1, dummy, 1, work, work.shape[0], 0)
            if result['info'] != 0:
                raise ValueError, 'Lapack SVD: ' + `result['info']`
            svmax = Numeric.maximum.reduce(sv)
            nvectors = Numeric.add.reduce(Numeric.greater(sv, 1.e-10*svmax))
            basis.shape = shape
            self._basis = ParticleVectorSet(self.universe, basis[:nvectors])
        return self._basis

    def projectionOf(self, vector):
        vector = vector.array
        basis = self.basis().array
        p = Numeric.zeros(vector.shape, Numeric.Float)
        for bv in basis:
            Numeric.add(Numeric.add.reduce(Numeric.ravel(bv*vector))*bv,
                        p, p)
        return Variables.ParticleVector(self.universe, p)

    def projectionComplementOf(self, vector):
        return vector - self.projectionOf(vector)


class RigidMotionSubspace(Subspace):

    def __init__(self, universe, objects):
        if not Utility.isSequenceObject(objects):
            objects = [objects]
        vectors = []
        for o in objects:
            atoms = o.atomList()
            for d in [Vector(1.,0.,0.), Vector(0.,1.,0.), Vector(0.,0.,1.)]:
                v = Variables.ParticleVector(universe)
                for a in atoms:
                    v[a] = d
                vectors.append(v)
            if len(atoms) > 1:
                center = o.centerOfMass()
                for d in [Vector(1.,0.,0.),Vector(0.,1.,0.),Vector(0.,0.,1.)]:
                    v = Variables.ParticleVector(universe)
                    for a in atoms:
                        v[a] = d.cross(a.position()-center)
                    vectors.append(v)
        Subspace.__init__(self, universe, vectors)


class PairDistanceSubspace(Subspace):

    def __init__(self, universe, atom_pairs):
        vectors = []
        for atom1, atom2 in atom_pairs:
            v = Variables.ParticleVector(universe)
            distance = atom1.position()-atom2.position()
            v[atom1] = distance
            v[atom2] = -distance
            vectors.append(v)
        Subspace.__init__(self, universe, vectors)
