# This module implements the various types of universes
# (infinite, periodic etc.). A universe defines the
# geometry of space, the force field, and external interactions
# (boundary conditions, external fields, etc.)
#
# Written by Konrad Hinsen
# last revision: 1999-3-10
#

import Bonds, ChemicalObjects, Collection, Environment, \
       Random, Utility, Variables, Visualization
from Scientific.Geometry import Transformation
from Scientific.Geometry import Vector, isVector
import copy, Numeric, operator, string

#
# The base class for all universes.
#
class Universe(Collection.GroupOfAtoms, Visualization.Viewable):

    def __init__(self, forcefield, properties):
	self._forcefield = forcefield
	self._evaluator = {}
	self.name = ''
	if properties.has_key('name'):
	    self.name = properties['name']
	    del properties['name']
	self._objects = Collection.Collection()
        self._environment = []
	self._configuration = None
	self.atom_properties = {}
        self._atoms = None
        self._bond_database = None
	self._version = 0
	self._np = None

    is_universe = 1
    is_periodic = 0

    def __getstate__(self):
	state = copy.copy(self.__dict__)
	state['_evaluator'] = {}
	state['_configuration'] = None
	del state['_bond_database']
	del state['_np']
        del state['_spec']
	return state

    def __setstate__(self, state):
	state['_np'] = None
	state['_atoms'] = None
	state['_bond_database'] = None
        self.__dict__['_environment'] = []
	for attr, value in state.items():
	    self.__dict__[attr] = value
	self._evaluator = {}
        self._createSpec()

    def __len__(self):
	return len(self._objects)

    def __getitem__(self, item):
	return self._objects[item]

    def __setattr__(self, attr, value):
	self.__dict__[attr] = value
	if attr[0] != '_':
	    self.addObject(value)

    def __delattr__(self, attr):
	self.removeObject(self.__dict__[attr])
	del self.__dict__[attr]

    def __repr__(self):
	return self.__class__.__name__ + ' ' + self.name + ' containing ' + \
               `len(self._objects)` + ' objects.'
    __str__ = __repr__

    def __copy__(self):
	return copy.deepcopy(self)

    def objectList(self):
	return self._objects.objectList()

    def atomList(self):
        if self._atoms is None:
            self._atoms = self._objects.atomList()
        return self._atoms

    def bondList(self):
	return self._objects.bondList()

    def danglingBondList(self):
	return self._objects.danglingBondList()

    def universe(self):
	return self

    def addObject(self, object):
	if Collection.isCollection(object):
	    for o in object:
		self.addObject(o)
	elif ChemicalObjects.isChemicalObject(object):
	    if object.parent is not None:
		if isUniverse(object.parent):
		    raise ValueError, `object` + \
			  ' is already in another universe'
		else:
		    raise ValueError, `object` + ' is part of another object'
	    object.parent = self
	    self._objects.addObject(object)
	    self._changed(1)
        elif Environment.isEnvironmentObject(object):
            for o in self._environment:
                o.checkCompatibilityWith(object)
            self._environment.append(object)
            self._changed(0)

    def removeObject(self, object):
	if Collection.isCollection(object):
	    for o in object:
		self.removeObject(o)
	elif ChemicalObjects.isChemicalObject(object):
	    if object.parent != self:
		raise ValueError, `object` + ' is not in this universe.'
	    object.parent = None
	    self._objects.removeObject(object)
	    self._changed(1)
        elif Environment.isEnvironmentObject(object):
            self._environment.remove(object)
	    self._changed(0)

    def selectShell(self, point, r1, r2=0.):
	return self._objects.selectShell(point, r1, r2)

    def selectBox(self, p1, p2):
	return self._objects.selectBox(p1, p2)

    def _changed(self, system_size_changed):
	self._evaluator = {}
        self._bond_database = None
	self._version = self._version + 1
        if (system_size_changed):
            self._configuration = None
            self.atom_properties = {}
            self._atoms = None
            self._np = None

    def setForceField(self, forcefield):
	self._forcefield = forcefield
	self._evaluator = {}
        self._bond_database = None

    def position(self, object, conf):
	if ChemicalObjects.isChemicalObject(object):
	    return object.position(conf)
	elif isVector(object):
	    return object
	else:
	    return Vector(object)

    def numberOfAtoms(self):
        return len(self.atomList())

    def numberOfPoints(self):
        if self._np is None:
            self._np = Collection.GroupOfAtoms.numberOfPoints(self)
        return self._np

    numberOfCartesianCoordinates = numberOfPoints

    def degreesOfFreedom(self):
	return 3*self.numberOfPoints()

    def configuration(self):
	if self._configuration is None:
	    np = self.numberOfPoints()
	    coordinates = Numeric.zeros((np, 3), Numeric.Float)
	    indices = range(np)
	    redef = []
	    for a in self._objects.atomList():
		if not a.setArray(coordinates, indices):
		    redef.append(a)
	    for a in redef:
		a.setArray(coordinates, indices[0:1])
		del indices[0]
	    self._configuration = 1
	    self._configuration = Variables.ParticleVector(self, coordinates)
            self._spec.foldCoordinatesIntoBox(coordinates)
	return self._configuration

    def atomNames(self):
	self.configuration()
	names = self.numberOfAtoms()*[None]
	for a in self._objects.atomList():
	    names[a.index] = a.fullName()
	return names

    def setConfiguration(self, configuration):
	conf = self.configuration()
	conf.assign(configuration)

    def _setConfiguration(self, coordinates):
	self._configuration = Variables.ParticleVector(self, coordinates)

    def addToConfiguration(self, displacement):
	conf = self.configuration()
	conf.assign(conf+displacement)

    def getAtomScalarArray(self, name, datatype = Numeric.Float):
	conf = self.configuration()
	array = Numeric.zeros((len(conf),), datatype)
	for a in self._objects.atomList():
	    array[a.index] = getattr(a, name)
	return Variables.ParticleScalar(self, array)

    def getAtomBooleanArray(self, name):
	conf = self.configuration()
	array = Numeric.zeros((len(conf),), Numeric.Int)
	for a in self._objects.atomList():
	    try:
		array[a.index] = getattr(a, name)
	    except AttributeError: pass
	return Variables.ParticleScalar(self, array)

    def renumberAtoms(self, indices):
	conf = self.configuration()
	for a in self.atomList():
	    a.index = indices[a.index]
	indices = Numeric.argsort(indices)
	conf.array[:] = Numeric.take(conf.array, indices)
	self._changed(1)

    def masses(self):
	return self.getAtomScalarArray('_mass')

    def velocities(self):
	try:
	    return self.atom_properties['velocity']
	except KeyError:
	    return None

    def setVelocities(self, velocities):
        try:
            v = self.atom_properties['velocity']
        except KeyError:
            v = Variables.ParticleVector(self)
            self.atom_properties['velocity'] = v
        v.assign(velocities)

    def initializeVelocitiesToTemperature(self, temperature):
	self.configuration()
	masses = self.masses()
	if self.atom_properties.has_key('velocity'):
	    del self.atom_properties['velocity']
	fixed = self.getAtomBooleanArray('fixed')
	np = self.numberOfPoints()
	velocities = Numeric.zeros((np, 3), Numeric.Float)
	for i in xrange(np):
	    m = masses[i]
	    if m > 0. and not fixed[i]:
		velocities[i] = Random.randomVelocity(temperature,
                                                           m).array
	self.atom_properties['velocity'] = \
			  Variables.ParticleVector(self, velocities)
        self.adjustVelocitiesToConstraints()

    def scaleVelocitiesToTemperature(self, temperature):
	velocities = self.velocities()
	factor = Numeric.sqrt(temperature/self.temperature())
	velocities.scaleBy(factor)

    def degreesOfFreedom(self):
        return GroupOfAtoms.degreesOfFreedom(self) \
               - self.numberOfDistanceConstraints()

    def distanceConstraintList(self):
        return self._objects.distanceConstraintList()

    def numberOfDistanceConstraints(self):
        return self._objects.numberOfDistanceConstraints()

    def setBondConstraints(self):
        self._objects.setBondConstraints(self)
        self.enforceConstraints()

    def removeDistanceConstraints(self):
        self._objects.removeDistanceConstraints(self)

    def enforceConstraints(self, configuration=None):
        import Dynamics
        Dynamics.enforceConstraints(self, configuration)
        self.adjustVelocitiesToConstraints()

    def adjustVelocitiesToConstraints(self, velocities=None):
        import Dynamics
        if velocities is None:
            velocities = self.velocities()
        if velocities is not None:
            Dynamics.projectVelocities(self, velocities)

    def bondLengthDatabase(self):
        if self._bond_database is None:
            self._bond_database = None
            if self._bond_database is None:
                self._bond_database = self._forcefield.bondLengthDatabase(self)
            if self._bond_database is None:
                self._bond_database = Bonds.DummyBondLengthDatabase(self)
        return self._bond_database

    def forcefield(self):
	return self._forcefield

    def energyEvaluator(self, subset1 = None, subset2 = None):
	if self._forcefield is None:
	    raise ValueError, "no force field defined"
	try:
	    eval = self._evaluator[(subset1, subset2)]
	except KeyError:
            from ForceFields import ForceField
	    eval = ForceField.EnergyEvaluator(self, self._forcefield,
					      subset1, subset2)
	    self._evaluator[(subset1, subset2)] = eval
	return eval

    def energy(self, subset1 = None, subset2 = None, small_change=0):
	eval = self.energyEvaluator(subset1, subset2)
	return eval(0, 0, small_change)

    def energyAndGradients(self, subset1 = None, subset2 = None,
			   small_change=0):
	eval = self.energyEvaluator(subset1, subset2)
	return eval(1, 0, small_change)

    def energyAndForceConstants(self, subset1 = None, subset2 = None,
				small_change=0):
	eval = self.energyEvaluator(subset1, subset2)
	e, g, fc =  eval(0, 1, small_change)
	return e, fc

    def energyGradientsAndForceConstants(self, subset1 = None, subset2 = None,
					 small_change=0):
	eval = self.energyEvaluator(subset1, subset2)
	return eval(1, 1, small_change)

    def energyTerms(self, subset1 = None, subset2 = None, small_change=0):
	eval = self.energyEvaluator(subset1, subset2)
	eval(0, 0, small_change)
	return eval.lastEnergyTerms()

    def distanceVector(self, p1, p2, conf=None):
	p1 = self.position(p1, conf)
	p2 = self.position(p2, conf)
        return Vector(self._spec.distanceVector(p1.array, p2.array))

    def distance(self, p1, p2, conf = None):
	return self.distanceVector(p1, p2, conf).length()

    def angle(self, p1, p2, p3, conf = None):
	v1 = self.distanceVector(p2, p1, conf)
	v2 = self.distanceVector(p2, p3, conf)
	return v1.angle(v2)

    def dihedral(self, p1, p2, p3, p4, conf = None):
	v1 = self.distanceVector(p2, p1, conf)
	v2 = self.distanceVector(p3, p2, conf)
	v3 = self.distanceVector(p3, p4, conf)
	a = v1.cross(v2).normal()
	b = v3.cross(v2).normal()
	cos = a*b
	sin = b.cross(a)*v2/v2.length()
	return Transformation.angleFromSineAndCosine(sin, cos)

    def _deleteAtom(self, atom):
	pass

    def _replaceAtom(self, atom, new_atom):
	pass

    def basisVectors(self):
	return None

    def reciprocalBasisVectors(self):
	return None

    def cellVolume(self):
	return None

    def randomPoint(self):
	raise TypeError, "undefined operation"

    def map(self, function):
	return self._objects.map(function)

    def description(self, objects = None, index_map = None):
	if objects is None:
	    objects = self
	return 'c(' + \
	       `self.__class__.__name__ + self._descriptionArguments()` + \
	       ',[' + string.join(map(lambda o, m=index_map: o.description(m),
				      objects.objectList()), ',') + ',])'

    def _graphics(self, conf, distance_fn, model, module, options):
	return self._objects._graphics(conf, distance_fn, model,
				       module, options)

    def setFromTrajectory(self, trajectory, step):
	conf = self.configuration()
	conf.assign(trajectory.configuration[step])

        volume = None
        try:
            volume = trajectory.volume[step]
        except AttributeError: pass
        if volume is not None:
            self.setVolume(volume)

	vel = self.velocities()
	try:
	    vel_tr = trajectory.velocities[step]
	except AttributeError:
	    if vel is not None:
		Utility.warning("velocities were not modified because the " +
				"trajectory does not contain velocity data.")
	    return
	if vel is None:
	    self.atom_properties['velocity'] = vel_tr
	else:
	    vel.assign(vel_tr)

    #
    # More efficient reimplementations of methods in Collection.GroupOfAtoms
    #
    def numberOfFixedAtoms(self):
        return self.getAtomBooleanArray('fixed').sumOverParticles()

    def degreesOfFreedom(self):
        return 3*(self.numberOfAtoms()-self.numberOfFixedAtoms()) \
               - self.numberOfDistanceConstraints()

    def mass(self):
        return self.masses().sumOverParticles()

    def centerOfMass(self, conf = None):
        m = self.masses()
        if conf is None:
            conf = self.configuration()
        return (m*conf).sumOverParticles()/m.sumOverParticles()

    def kineticEnergy(self, velocities = None):
	if velocities is None:
	    velocities = self.velocities()
        return 0.5*velocities.massWeightedDotProduct(velocities)

    def momentum(self, velocities = None):
	if velocities is None:
	    velocities = self.velocities()
        return (self.masses()*velocities).sumOverParticles()

#
# Infinite universes
#
class InfiniteUniverse(Universe):

    def __init__(self, forcefield=None, **properties):
	Universe.__init__(self, forcefield, properties)
        self._createSpec()

##      def distanceVector(self, p1, p2, conf = None):
##  	p1 = self.position(p1, conf)
##  	p2 = self.position(p2, conf)
##  	return p2-p1

    def realToBoxCoordinates(self, vector):
        return vector
    
    def boxToRealCoordinates(self, vector):
        return vector

    def CdistanceFunction(self):
	from MMTK_universe import infinite_universe_distance_function
	return infinite_universe_distance_function, Numeric.array([0.])

    def CcorrectionFunction(self):
	from MMTK_universe import infinite_universe_correction_function
	return infinite_universe_correction_function, Numeric.array([0.])

    def CvolumeFunction(self):
	from MMTK_universe import infinite_universe_volume_function
	return infinite_universe_volume_function, Numeric.array([0.])

    def CboxTransformationFunction(self):
        return None, Numeric.array([0.])

    def _createSpec(self):
        from MMTK_universe import InfiniteUniverseSpec
        self._spec = InfiniteUniverseSpec()

    def _descriptionArguments(self):
	return '()'

#
# Orthorhombic universe with periodic boundary conditions
#
class OrthorhombicPeriodicUniverse(Universe):

    def __init__(self, size = None, forcefield = None, **properties):
	Universe.__init__(self, forcefield, properties)
	self.data = Numeric.zeros((2, 3), Numeric.Float)
	self.edges = self.data[0] # for compatibility with previous versions
	if size is not None:
	    self.setSize(size)
	Universe.__init__(self, forcefield, properties)
        self._createSpec()

    is_periodic = 1

    def __setstate__(self, state):
        Universe.__setstate__(self, state)
        self.edges = self.data[0]

    def setSize(self, size):
	self.data[0, 0], self.data[0, 1], self.data[0, 2] = size
	self.data[1] = 0.5*self.data[0]

    def scaleSize(self, factor):
	self.data[:] = factor*self.data

    def setVolume(self, volume):
        factor = (volume/self.cellVolume())**(1./3.)
        self.scaleSize(factor)

##      def distanceVector(self, p1, p2, conf = None):
##  	p1 = self.position(p1, conf)
##  	p2 = self.position(p2, conf)
##  	d = p2-p1
##  	edge = self.data[0, 0]
##  	half_edge = 0.5*edge
##  	if d[0] < 0.: half_edge = -half_edge
##  	x = Numeric.fmod(d[0]+half_edge, edge) - half_edge
##  	edge = self.data[0, 1]
##  	half_edge = 0.5*edge
##  	if d[1] < 0.: half_edge = -half_edge
##  	y = Numeric.fmod(d[1]+half_edge, edge) - half_edge
##  	edge = self.data[0, 2]
##  	half_edge = 0.5*edge
##  	if d[2] < 0.: half_edge = -half_edge
##  	z = Numeric.fmod(d[2]+half_edge, edge) - half_edge
##  	return Vector(x,y,z)

    def realToBoxCoordinates(self, vector):
        x, y, z = vector
        return Vector(x/self.data[0, 0],
                      y/self.data[0, 1],
                      z/self.data[0, 2])

    def boxToRealCoordinates(self, vector):
        x, y, z = vector
        return Vector(x*self.data[0, 0],
                      y*self.data[0, 1],
                      z*self.data[0, 2])

    def CdistanceFunction(self):
	from MMTK_universe import orthorhombic_universe_distance_function
	return orthorhombic_universe_distance_function, self.data

    def CcorrectionFunction(self):
	from MMTK_universe import orthorhombic_universe_correction_function
	return orthorhombic_universe_correction_function, self.data

    def CvolumeFunction(self):
	from MMTK_universe import orthorhombic_universe_volume_function
	return orthorhombic_universe_volume_function, self.data

    def CboxTransformationFunction(self):
	from MMTK_universe import orthorhombic_universe_box_transformation
	return orthorhombic_universe_box_transformation, self.data

    def basisVectors(self):
	return [self.data[0, 0]*Vector(1., 0., 0.),
                self.data[0, 1]*Vector(0., 1., 0.),
                self.data[0, 2]*Vector(0., 0., 1.)]

    def reciprocalBasisVectors(self):
	return [Vector(1., 0., 0.)/self.data[0, 0],
                Vector(0., 1., 0.)/self.data[0, 1],
                Vector(0., 0., 1.)/self.data[0, 2]]

    def cellVolume(self):
	return Numeric.multiply.reduce(self.data[0])

    def randomPoint(self):
	return Random.randomPointInBox(self.data[0, 0], self.data[0, 1],
					    self.data[0, 2])

    def _createSpec(self):
        from MMTK_universe import OrthorhombicPeriodicUniverseSpec
        self._spec = OrthorhombicPeriodicUniverseSpec(self.data)

    def _descriptionArguments(self):
	return `((self.data[0, 0], self.data[0, 1], self.data[0, 2]),)`

#
# Cubic universe with periodic boundary conditions
#
class CubicPeriodicUniverse(OrthorhombicPeriodicUniverse):

    def setSize(self, size):
	OrthorhombicPeriodicUniverse.setSize(self, 3*(size,))

    def _descriptionArguments(self):
	return `(self.data[0, 0],)`

#
# Recognition functions
#
def isUniverse(object):
    return hasattr(object, 'is_universe')
