# This module contains interfaces to external visualization programs
# and a visualization base class
#
# Written by Konrad Hinsen
# last revision: 1999-2-19
#

import ConfigIO, PDB, Universe, Utility
import Numeric, string, tempfile, os

#
# If you want temporary files in a non-standard directory, make
# its name the value of this variable:
#
tempdir = None

#
# Get visualization program names
#
viewer = {}
prog = None
try:
    viewer['pdb'] = os.environ['PDBVIEWER']
    prog = os.path.split(string.split(viewer['pdb'])[0])[1]
except KeyError: pass
try:
    viewer['vrml'] = os.environ['VRMLVIEWER']
except KeyError: pass

#
# Visualization base class. Defines methods for general visualization
# tasks.
#
class Viewable:

    def graphicsObjects(self, **options):
        conf = options.get('configuration', None)
        model = options.get('model', 'wireframe')
	try:
	    module = options['graphics_module']
	except KeyError:
	    import VRML
	    module = VRML
        color_values = options.get('color_values', None)
        if color_values is not None:
            lower = Numeric.minimum.reduce(color_values.array)
            upper = Numeric.maximum.reduce(color_values.array)
            options['color_scale'] = module.ColorScale((lower, upper))
	try:
	    distance_fn = self.universe().distanceVector
	except AttributeError:
	    distance_fn = Universe.InfiniteUniverse().distanceVector
	return self._graphics(conf, distance_fn, model, module, options)

    def _atomColor(self, atom, options):
        color_values = options.get('color_values', None)
        if color_values is None:
            return atom.color
        else:
            scale = options['color_scale']
            return scale(color_values[atom])

#
# View anything viewable.
#
def view(object, *parameters):
    apply(object.view, parameters)

#
# Display an object or a collection of objects using a
# PBD viewing program
#
def genericViewConfiguration(object, configuration = None, format = 'pdb'):
    if len(viewer) == 0:
	Utility.warning('No PDB or VRML viewer defined.')
	return
    format = string.lower(format)
    viewer_format = string.split(format, '.')[0]
    if not viewer.has_key(viewer_format):
	format = viewer.keys()[0]
	viewer_format = string.split(format, '.')[0]

    tempfile.tempdir = tempdir
    filename = tempfile.mktemp()
    tempfile.tempdir = None
    file = ConfigIO.OutputFile(filename, format)
    file.write(object, configuration)
    file.close()
    if os.fork() == 0:
	pipe = os.popen(viewer[viewer_format] + ' ' + filename + \
			' 1> /dev/null 2>&1', 'w')
	pipe.close()
	os.unlink(filename)
	os._exit(0)

viewConfiguration = genericViewConfiguration

#
# Normal mode and trajectory animation: in general not available
#
def viewSequence(*args):
    Utility.warning('No viewer with animation feature defined.')

def viewTrajectory(trajectory, first=0, last=None, step=1, subset = None):
    if type(trajectory) == type(''):
	from Trajectory import Trajectory
	trajectory = Trajectory(None, trajectory, 'r')
    if last is None:
	last = len(trajectory)
    elif last < 0:
	last = len(trajectory) + last
    universe = trajectory.universe
    if subset is None:
	subset = universe
    viewSequence(subset, trajectory.configuration[first:last:step])
    
def viewMode(mode, factor=1., subset=None):
    universe = mode.universe
    if subset is None:
	subset = universe
    conf = universe.configuration()
    viewSequence(subset, [conf, conf+factor*mode, conf, conf-factor*mode], 1)

#
# XMol support
#    
if prog == 'xmol':
    #
    # Animation with XMol.
    #
    def viewSequence(object, conf_list, periodic = 0):
	tempfile.tempdir = tempdir
	file_list = []
	for conf in conf_list:
	    file = tempfile.mktemp()
	    file_list.append(file)
	    object.writeToFile(file, conf, 'pdb')
	bigfile = tempfile.mktemp()
	tempfile.tempdir = None
	os.system('cat ' + string.join(file_list, ' ') + ' > ' + bigfile)
	for file in file_list:
	    os.unlink(file)
	if os.fork() == 0:
	    pipe = os.popen('xmol -readFormat pdb ' + bigfile + \
			    ' 1> /dev/null 2>&1', 'w')
	    pipe.close()
	    os.unlink(bigfile)
	    os._exit(0)

#
# VMD support
#
if prog == 'vmd':
    #
    # View configuration
    #
    def viewConfiguration(object, configuration = None, format = 'pdb'):
	format = string.lower(format)
	if format != 'pdb':
	    return genericViewConfiguration(object, configuration, format)
	tempfile.tempdir = tempdir
	filename = tempfile.mktemp()
	script = tempfile.mktemp()
	tempfile.tempdir = None
	file = ConfigIO.OutputFile(filename, format)
	file.write(object, configuration)
	file.close()
	file = open(script, 'w')
	file.write('mol load pdb ' + filename + '\n')
	file.write('color Name 1 white\n')
	file.write('color Name 2 white\n')
	file.write('color Name 3 white\n')
	file.write('file delete ' + filename + '\n')
	file.write('file delete ' + script + '\n')
	file.close()
	os.system(viewer[format] + ' -nt -e ' + script + ' 1> /dev/null 2>&1')
    #
    # Animate sequence
    #
    def viewSequence(object, conf_list, periodic = 0):
	tempfile.tempdir = tempdir
	script = tempfile.mktemp()
	np = object.numberOfPoints()
	universe = object.universe()
	if np == universe.numberOfPoints() \
	   and len(conf_list) > 2:
	    import DCD
	    pdbfile = tempfile.mktemp()
	    dcdfile = tempfile.mktemp()
	    tempfile.tempdir = None
	    universe.configuration()
	    pdb = PDB.PDBOutputFile(pdbfile)
	    pdb.write(object, conf_list[0])
	    indices = map(lambda a: a.index, pdb.atom_sequence)
	    pdb.close()
	    del pdb
	    DCD.writeDCD(conf_list[1:], dcdfile, indices)
	    file = open(script, 'w')
	    file.write('mol load pdb ' + pdbfile + '\n')
	    file.write('animate read dcd ' + dcdfile + '\n')
	    if periodic:
		file.write('animate style loop\n')
	    else:
		file.write('animate style once\n')
	    file.write('animate forward\n')
	    file.write('file delete ' + pdbfile + '\n')
	    file.write('file delete ' + dcdfile + '\n')
	    file.write('file delete ' + script + '\n')
	    file.close()
	else:
	    file_list = []
	    for conf in conf_list:
		file = tempfile.mktemp()
		file_list.append(file)
		object.writeToFile(file, conf, 'pdb')
	    tempfile.tempdir = None
	    file = open(script, 'w')
	    file.write('mol load pdb ' + file_list[0] + '\n')
	    for conf in file_list[1:]:
		file.write('animate read pdb ' + conf + '\n')
	    if periodic:
		file.write('animate style loop\n')
	    else:
		file.write('animate style once\n')
	    file.write('animate forward\n')
	    for conf in file_list:
		file.write('file delete ' + conf + '\n')
	    file.write('file delete ' + script + '\n')
	    file.close()
	os.system(viewer['pdb'] + ' -nt -e ' + script + ' 1> /dev/null 2>&1')

