/* I/O in DCD format
 *
 * Written by Lutz Ehrlich
 * Adapted to MMTK conventions by Konrad Hinsen
 * last revision: 1999-3-19
 */

#include "mmtk.h"
#include "mmtk_forcefield.h"
#include "mmtk_trajectory.h"
#include "mmtk_readdcd.h"

/* Global variables */

double angstrom_factor;
double akma_time_factor;

/* Allocate and initialize Output variable descriptors */

static  PyTrajectoryVariable *
get_data_descriptors(PyArrayObject *configuration, double *time)
{
  static PyTrajectoryVariable vars[3];
  int i;
  if (vars != NULL) {
    vars[0].name = "time";
    vars[0].text = "Time: %lf\n";
    vars[0].unit = time_unit_name;
    vars[0].type = PyTrajectory_Scalar;
    vars[0].class = PyTrajectory_Time;
    vars[0].value.dp = time;
    vars[1].name = "configuration";
    vars[1].text = "Configuration:\n";
    vars[1].unit = length_unit_name;
    vars[1].type = PyTrajectory_ParticleVector;
    vars[1].class = PyTrajectory_Configuration;
    vars[1].value.array = configuration;
    vars[2].name = NULL;
  }
  return vars;
}


static  PyObject *
readDCD(PyObject *dummy, PyObject *args)
{
  PyObject *universe;
  PyArrayObject *configuration;
  PyListObject *spec_list;
  PyTrajectoryOutputSpec *output;
  vector3 *x;
  int atoms;

  char buffer[100];
  PyTrajectoryVariable *data_descriptors;
  int i, j;

  /* DCD transformation variables */
  char * dcdFileName;
  int    dcdFileNum;
  int *  dcdFreeatoms;
  int    dcdAtoms;
  int    dcdNFrames=0;
  int    dcdFrameStart=0;
  int    dcdFrameSkip=0;
  float  dcdTimeStep;
  int    dcdNamnf;
  float *dcdX = NULL;
  float *dcdY = NULL;
  float *dcdZ = NULL;
  int    dcdErrcode;
  int    currFrames=0;
  double time;

  dcdFreeatoms = NULL;
    

  /* Parse and check arguments */
  if (!PyArg_ParseTuple(args, "OO!O!s", &universe,
			&PyArray_Type, &configuration,
			&PyList_Type, &spec_list, &dcdFileName))
    return NULL;

  atoms = configuration->dimensions[0];
  x = (vector3 *)configuration->data;

  /* Prepare output data descriptors */
  data_descriptors = get_data_descriptors(configuration, &time);

  /* Initialize output */
  output = PyTrajectory_OutputSpecification(universe, spec_list,
					    dcdFileName,
					    data_descriptors);
  if (output == NULL)
    return NULL;

  /* open DCD file  */
  dcdFileNum = open_dcd_read( dcdFileName );
  if ( dcdFileNum < 0 ){
    PyErr_SetString(PyExc_IOError, "Cannot open file");
    goto error;
  }

  /* read header */
  dcdErrcode = read_dcdheader(dcdFileNum, &dcdAtoms, &dcdNFrames, 
			      &dcdFrameStart,  &dcdFrameSkip, &dcdTimeStep,
			      &dcdNamnf, &dcdFreeatoms);
  if( atoms != dcdAtoms ){
    sprintf(buffer, "number of atoms in DCD file (%d) doesn't "
	    "match universe (%d)", dcdAtoms, atoms);
    PyErr_SetString(PyExc_ValueError, buffer);
    goto error;
  }
  if ( dcdNamnf != 0 ){
    PyErr_SetString(PyExc_ValueError, "Can't read DCD files with free atoms");
    goto error;
  }

  /* allocate the dcd{X,Y,Z} arrays */
  dcdX = ( float*) malloc(dcdAtoms * sizeof(float) );
  dcdY = ( float*) malloc(dcdAtoms * sizeof(float) );
  dcdZ = ( float*) malloc(dcdAtoms * sizeof(float) );
  if( (dcdX==NULL) || (dcdY==NULL) || (dcdZ==NULL) ){
    PyErr_NoMemory();
    goto error;
  }
  
  /* read in the frames one after the other */
  currFrames = 0;
  time = 0.;
  while ( 1 ){
    int err_code = read_dcdstep(dcdFileNum, dcdAtoms, dcdX, dcdY, dcdZ,
				dcdNamnf, (currFrames == 0), dcdFreeatoms);
    if (err_code == -1)
      break;
    if (err_code < 0) {
      PyErr_SetString(PyExc_IOError, "DCD read error");
      goto error;
    }
    for (j = 0; j < dcdAtoms; j++) {
       x[j][0] = angstrom_factor * dcdX[j];
       x[j][1] = angstrom_factor * dcdY[j];
       x[j][2] = angstrom_factor * dcdZ[j];
    }
    if (PyTrajectory_Output(output, currFrames, data_descriptors) == -1)
      goto error;
    currFrames++;
    time += dcdFrameSkip*dcdTimeStep*akma_time_factor;
  }
  close_dcd_read(dcdFileNum,0,dcdFreeatoms);

  /* Clean up and return None */
  if( dcdX != NULL )
    free(dcdX);
  if( dcdY != NULL )
    free(dcdY);
  if( dcdZ != NULL )
    free(dcdZ);
  PyTrajectory_OutputFinish(output, currFrames-1, 0, data_descriptors);
  Py_INCREF(Py_None);
  return Py_None;

  /* Clean up and return error */
error:
  if( dcdX != NULL )
    free(dcdX);
  if( dcdY != NULL )
    free(dcdY);
  if( dcdZ != NULL )
    free(dcdZ);
  close_dcd_read(dcdFileNum,0,dcdFreeatoms);
  PyTrajectory_OutputFinish(output, currFrames, 1, data_descriptors);
  return NULL;
}


static PyObject *
writeOpenDCD(PyObject *dummy, PyObject *args)
{
  PyObject *result;

  /* DCD transformation variables */
  char * dcdFileName;
  int    dcdAtoms;
  int    dcdNFrames=0;
  int    dcdFrameStart=0;
  int    dcdNSavc;
  int    dcdErrcode;
  int    currFrames=0;
  double time;
  int fd;
  int ii;

  /* Parse and check arguments */
  if (!PyArg_ParseTuple(args, "siiiid", &dcdFileName,&dcdAtoms, &dcdNFrames, 
			&dcdFrameStart, &dcdNSavc, &time))
    return NULL;

  /* open DCD file  */
  fd = open_dcd_write( dcdFileName );
  if ( fd == DCD_OPENFAILED ){
    PyErr_SetString(PyExc_IOError, "Cannot open file");
    return NULL;
  }


  /* write header */
  dcdErrcode = write_dcdheader(fd , dcdFileName, dcdAtoms, dcdNFrames, 
			      dcdFrameStart,  dcdNSavc, time);
  result = (PyObject*)PyInt_FromLong((long) fd);  /* ugly, ugly, ugly */
  return result;

}

static  PyObject *
writeDCDStep(PyObject *dummy, PyObject *args)
{
  int i, j;
  /* DCD transformation variables */
  char * dcdFileName;
  int    dcdAtoms;
  float *dcdX = NULL;
  float *dcdY = NULL;
  float *dcdZ = NULL;
  int err;
  int fd;
  int ii;
  PyArrayObject *config, *atom_order;
  vector3 *pos;
  long *index;

  /* Parse and check arguments */
  if (!PyArg_ParseTuple(args, "iO!O!", &fd,
			&PyArray_Type, &config,
			&PyArray_Type, &atom_order))
    return NULL;
  
  dcdAtoms = config->dimensions[0];
  pos      = (vector3*) config->data;
  index    = (long *) atom_order->data;
  /* allocate the dcd{X,Y,Z} arrays */
  
  dcdX = ( float*) malloc(dcdAtoms * sizeof(float) );
  dcdY = ( float*) malloc(dcdAtoms * sizeof(float) );
  dcdZ = ( float*) malloc(dcdAtoms * sizeof(float) );
  if( (dcdX==NULL) || (dcdY==NULL) || (dcdZ==NULL) ){
    PyErr_NoMemory();
    goto error;
  }
  for( ii=0; ii<dcdAtoms; ii++ ){
    int jj = index[ii];
    dcdX[ii] = (float) pos[jj][0] / angstrom_factor;
    dcdY[ii] = (float) pos[jj][1] / angstrom_factor;
    dcdZ[ii] = (float) pos[jj][2] / angstrom_factor;
  }
    

  err = write_dcdstep( fd, dcdAtoms, dcdX, dcdY, dcdZ);
  if( err!=1 ){
    printf("Couldn't write DCD step\n");
    goto error;
  }

  if ( dcdX ) free(dcdX);
  if ( dcdY ) free(dcdY);
  if ( dcdZ ) free(dcdZ);

  return Py_None;

  /* Clean up and return error */
error:
  if ( dcdX ) free(dcdX);
  if ( dcdY ) free(dcdY);
  if ( dcdZ ) free(dcdZ);
  return NULL;
}

static  PyObject *
writeCloseDCD(PyObject *dummy, PyObject *args)
{
  int fd;

  /* Parse and check arguments */
  if (!PyArg_ParseTuple(args, "i",&fd))
    return NULL;

  close_dcd_read(fd, 0, NULL);

  return Py_None;

}


static PyMethodDef DCD_methods[] = {
  {"readDCD", readDCD, 1},
  {"writeOpenDCD", writeOpenDCD, 1},
  {"writeDCDStep", writeDCDStep, 1},
  {"writeCloseDCD", writeCloseDCD, 1},
  {NULL, NULL}		/* sentinel */
};


/* Initialization function for the module */

void
initMMTK_DCD()
{
  PyObject *m;
  PyObject *trajectory, *forcefield, *units;

  /* Create the module and add the functions */
  m = Py_InitModule("MMTK_DCD", DCD_methods);

  /* Import the array module */
#ifdef import_array
  import_array();
#endif
  
  /* Import _forcefield and _trajectory and retrieve C API address arrays */
  forcefield = PyImport_ImportModule("MMTK_forcefield");
  if (forcefield != NULL) {
    PyObject *module_dict = PyModule_GetDict(forcefield);
    PyObject *c_api_object = PyDict_GetItemString(module_dict, "_C_API");
    if (PyCObject_Check(c_api_object)) {
      PyFF_API = (void **)PyCObject_AsVoidPtr(c_api_object);
    }
  }
  trajectory = PyImport_ImportModule("MMTK_trajectory");
  if (trajectory != NULL) {
    PyObject *module_dict = PyModule_GetDict(trajectory);
    PyObject *c_api_object = PyDict_GetItemString(module_dict, "_C_API");
    if (PyCObject_Check(c_api_object)) {
      PyTrajectory_API = (void **)PyCObject_AsVoidPtr(c_api_object);
    }
  }

  /* Get the length and time unit conversion factor from Units */
  units = PyImport_ImportModule("MMTK.Units");
  if (units != NULL) {
    PyObject *module_dict = PyModule_GetDict(units);
    PyObject *factor = PyDict_GetItemString(module_dict, "Ang");
    angstrom_factor = PyFloat_AsDouble(factor);
    factor = PyDict_GetItemString(module_dict, "akma_time");
    akma_time_factor = PyFloat_AsDouble(factor);
  }

  /* Check for errors */
  if (PyErr_Occurred())
    Py_FatalError("can't initialize module MMTK_DCD");
}
