/* Ewald method for electrostatic interactions
 *
 * Written by Konrad Hinsen
 * last revision: 1999-3-23
 */


#define NO_IMPORT
#define _FORCEFIELD_MODULE
#include "mmtk_forcefield.h"
#include "mmtk_forcefield_private.h"

/*
 * Include erfc code, unless user wants to use an erfc from libm.
 */
#ifndef LIBM_HAS_ERFC
#include "polevl.c"
#include "ndtr.c"
#endif

double
es_ewald_evaluator(PyObject *object, int natoms,
		   PyArrayObject *coordinates,
		   PyObject *gradients, PyObject *force_constants,
		   gradient_function *gradient_fn, fc_function *fc_fn,
		   int small_change)
{
  PyFFEvaluatorObject *self = (PyFFEvaluatorObject *)object;
  vector3 *x = (vector3 *)coordinates->data;
  distance_fn *d_fn = self->universe_spec->distance_function;
  box_fn *box_transformation_fn = self->universe_spec->box_function;
  volume_fn *v_fn = self->universe_spec->volume_function;
  double *universe_data = self->universe_spec->geometry_data;
  double volume = (*v_fn)(1., universe_data);

  PyNonbondedListObject *nblist = (PyNonbondedListObject *)self->data1;
  double *charge = (double *)((PyArrayObject *)self->data2)->data;
  long *kmax = (long *)((PyArrayObject *)self->data3)->data;
  double cutoff_sq = sqr(self->param1);
  double one_four_factor = self->param2;
  double beta = self->param3;
  double reciprocal_cutoff_sq = self->param4;
  atom_pair *pair = nblist->pairs;
  long *excluded = (long *)((PyArrayObject *)nblist->excluded_pairs)->data;
  int n_ex = 2*((PyArrayObject *)nblist->excluded_pairs)->dimensions[0];
  double charge_sum;
  int k;

  self->energy = 0.;

  if (d_fn == orthorhombic_distance_vector_pointer) {
#   define distance_vector_function distance_vector_2
#   include "ewald1.i"
#   undef distance_vector_function
  }
  else {
#   define distance_vector_function (*d_fn)
#   include "ewald1.i"
#   undef distance_vector_function
  }

  charge_sum = 0.;
  for (k = 0; k < natoms; k++)
    charge_sum += charge[k]*charge[k];
  self->energy -= beta*charge_sum*electrostatic_energy_factor/sqrt(M_PI);

  self->energy += reciprocal_sum(natoms, x, volume, charge, beta,
				 kmax, reciprocal_cutoff_sq,
				 box_transformation_fn, universe_data,
				 gradients, gradient_fn,
				 force_constants, fc_fn, self,
				 self->scratch);

  self->virial = self->energy;
  self->virial_available = 1;
  return self->energy;
}

/*
 * Plain Ewald sum, for orthorhombic systems only.
 */
typedef struct {
   double real;
   double imag;
} complex;

static complex c_1 = {1., 0.};

static complex
c_mult(complex a, complex b)
{
  complex r;
  r.real = a.real*b.real - a.imag*b.imag;
  r.imag = a.real*b.imag + a.imag*b.real;
  return r;
}

static double
c_abssq(complex z)
{
  return z.real*z.real + z.imag*z.imag;
}

static double
reciprocal_sum(int natoms, vector3 *x, double volume,
	       double *charge, double beta,
	       long *kmax, double cutoff_sq,
	       box_fn *box_transformation_fn, double *universe_data,
	       PyObject *gradients, gradient_function *gradient_fn,
	       PyObject *force_constants, fc_function *fc_fn,
	       PyFFEvaluatorObject *self,
	       void *scratch)
{
  vector3 *xb = (vector3 *)scratch;
  complex *eikx = (complex *)(xb + natoms);
  complex *eiky = eikx + natoms*(kmax[0]+1);
  complex *eikz = eiky + natoms*(2*kmax[1]+1);
  complex *eikr = eikz + natoms*(2*kmax[2]+1);
  int *nkvect = (int *)(eikr+natoms);
  int *kx = nkvect + 1;
  int *ky = kx + *nkvect;
  int *kz = ky + *nkvect;
  vector3 r1, r2, r3;
  int nx, ny, nz, negy, negz, xk, yk, zk;
  double energy;
  int i, nk;

  r1[0] = 2.*M_PI;  r1[1] = r1[2] = 0.;
  r2[1] = 2.*M_PI;  r2[2] = r2[0] = 0.;
  r3[2] = 2.*M_PI;  r3[0] = r3[1] = 0.;
  (*box_transformation_fn)(&r1, &r1, 1, universe_data, 1);
  (*box_transformation_fn)(&r2, &r2, 1, universe_data, 1);
  (*box_transformation_fn)(&r3, &r3, 1, universe_data, 1);

  nx = kmax[0]+1;
  ny = 2*kmax[1]+1;
  nz = 2*kmax[2]+1;
  negy = kmax[1];
  negz = kmax[2];

  (*box_transformation_fn)(x, xb, natoms, universe_data, 1);
  for (i = 0; i < natoms; i++) {
    xb[i][0] *= 2.*M_PI;
    xb[i][1] *= 2.*M_PI;
    xb[i][2] *= 2.*M_PI;
  }

  for (i = 0; i < natoms; i++) {
    eikx[nx*i] = c_1;
    eiky[ny*i] = c_1;
    eikz[nz*i] = c_1;
    eikx[nx*i+1].real = cos(xb[i][0]);
    eikx[nx*i+1].imag = sin(xb[i][0]);
    eiky[ny*i+1].real = cos(xb[i][1]);
    eiky[ny*i+1].imag = sin(xb[i][1]);
    eikz[nz*i+1].real = cos(xb[i][2]);
    eikz[nz*i+1].imag = sin(xb[i][2]);
    eiky[ny*i+negy+1].real = eiky[ny*i+1].real;
    eiky[ny*i+negy+1].imag = -eiky[ny*i+1].imag;
    eikz[nz*i+negz+1].real = eikz[nz*i+1].real;
    eikz[nz*i+negz+1].imag = -eikz[nz*i+1].imag;
  }
  for (xk = 2; xk <= kmax[0]; xk++)
    for (i = 0; i < natoms; i++)
      eikx[nx*i+xk] = c_mult(eikx[nx*i+xk-1], eikx[nx*i+1]);
  for (yk = 2; yk <= kmax[1]; yk++)
    for (i = 0; i < natoms; i++) {
      eiky[ny*i+yk] = c_mult(eiky[ny*i+yk-1], eiky[ny*i+1]);
      eiky[ny*i+negy+yk].real = eiky[ny*i+yk].real;
      eiky[ny*i+negy+yk].imag = -eiky[ny*i+yk].imag;
    }
  for (zk = 2; zk <= kmax[2]; zk++)
    for (i = 0; i < natoms; i++) {
      eikz[nz*i+zk] = c_mult(eikz[nz*i+zk-1], eikz[nz*i+1]);
      eikz[nz*i+negz+zk].real = eikz[nz*i+zk].real;
      eikz[nz*i+negz+zk].imag = -eikz[nz*i+zk].imag;
    }

  energy = 0.;
  for (nk = 0; nk < *nkvect; nk++) {
    int xk = kx[nk];
    int yk = ky[nk];
    int zk = kz[nk];
    int iyk = (yk < 0) ? negy-yk : yk;
    int izk = (zk < 0) ? negz-zk : zk;
    vector3 k = {0., 0., 0.};
    double ksq, expfactor;
    complex sum;

    vector_add(k, r1, xk);
    vector_add(k, r2, yk);
    vector_add(k, r3, zk);
    ksq = vector_length_sq(k);
    expfactor = exp(-ksq/(4.*beta*beta))/ksq;
    if (kx != 0)
      expfactor *= 2;

    sum.real = sum.imag = 0.;
    for (i = 0; i < natoms; i++) {
      eikr[i] = c_mult(c_mult(eikx[nx*i+xk], eiky[ny*i+iyk]),
		       eikz[nz*i+izk]);
      sum.real += charge[i]*eikr[i].real;
      sum.imag += charge[i]*eikr[i].imag;
    }
    energy += c_abssq(sum)*expfactor;
    if (gradients != NULL) {
      for (i = 0; i < natoms; i++) {
	vector3 grad;
	double a = 4.*M_PI*expfactor*electrostatic_energy_factor *
	  charge[i]*(sum.imag*eikr[i].real-sum.real*eikr[i].imag)/volume;
	grad[0] = a*k[0];
	grad[1] = a*k[1];
	grad[2] = a*k[2];
	if (gradient_fn != NULL)
	  (*gradient_fn)(self, gradients, i, grad);
	else {
	  vector3 *f = (vector3 *)((PyArrayObject *)gradients)->data;
	  f[i][0] += grad[0];
	  f[i][1] += grad[1];
	  f[i][2] += grad[2];
	}
      }
    }
    if (force_constants != NULL) {
      tensor3 fij;
      int j;
      tensor_product(fij, k, k,
		     4.*M_PI*expfactor*electrostatic_energy_factor/volume);
      for (i = 0; i < natoms; i++)
	for (j = i; j < natoms; j++) {
	  double f = charge[i]*charge[j];
	}
      if (fc_fn != NULL) {
      }
      else {
      }
    }
  }

  return 2.*M_PI*energy*electrostatic_energy_factor/volume;
}

/*
 * Count k vectors for memory allocation and initialize scratch area.
 */
int
init_kvectors(box_fn *box_transformation_fn, double *universe_data, int natoms,
	      long *kmax, double cutoff_sq, void *scratch, int nvect)
{
  vector3 *xb = (vector3 *)scratch;  /* Size natoms */
  complex *eikx = (complex *)(xb + natoms);  /* Size (kmax[0]+1)*natoms */
  complex *eiky = eikx + natoms*(kmax[0]+1);  /* Size (2*kmax[1]+1)*natoms */
  complex *eikz = eiky + natoms*(2*kmax[1]+1);  /* Size (2*kmax[2]+1)*natoms */
  complex *eikr = eikz + natoms*(2*kmax[2]+1);  /* Size natoms */
  int *nkvect = (int *)(eikr+natoms);  /* Size 1 */
  int *kx = nkvect + 1;  /* Size nvect */
  int *ky = kx + nvect;  /* Size nvect */
  int *kz = ky + nvect;  /* Size nvect */

  vector3 r1, r2, r3, k;
  int x, y, z;
  int nk = 0;

  r1[0] = 2.*M_PI;  r1[1] = r1[2] = 0.;
  r2[1] = 2.*M_PI;  r2[2] = r2[0] = 0.;
  r3[2] = 2.*M_PI;  r3[0] = r3[1] = 0.;
  (*box_transformation_fn)(&r1, &r1, 1, universe_data, 1);
  (*box_transformation_fn)(&r2, &r2, 1, universe_data, 1);
  (*box_transformation_fn)(&r3, &r3, 1, universe_data, 1);

  if (scratch != NULL)
    *nkvect = nvect;
  for (x = 0; x <= kmax[0]; x++) {
    for (y = -kmax[1]; y <= kmax[1]; y++) {
      for (z = -kmax[2]; z <= kmax[2]; z++) {
	vector3 k = {0., 0., 0.};
	double ksq;
	vector_add(k, r1, x);
	vector_add(k, r2, y);
	vector_add(k, r3, z);
	ksq = vector_length_sq(k);
	if (ksq < cutoff_sq && ksq > 0) {
	  if (scratch != NULL) {
	    kx[nk] = x; ky[nk] = y; kz[nk] = z;
	  }
	  nk++;
	}
      }
    }
  }
#if 0
  printf("%d k vectors\n", nk);
#endif
  return nk;
}
