/* Low-level force field calculations: non-bonded interactions
 *
 * Written by Konrad Hinsen
 * last revision: 1999-3-17
 */

#define NO_IMPORT
#define _FORCEFIELD_MODULE
#include "mmtk_forcefield.h"
#include "mmtk_forcefield_private.h"

/* PMTA definitions */

#ifdef WITH_DPMTA
#include "dpmta.h"
#endif

/* Factor by which cutoffs are multiplied to provide a boundary
   region in which changes are monitored */
const double safety_factor = 1.08;
const double check_limit = 1.06;

/* Number of iterations before the non-bonded list is always remade */
const int remake_frequency = 50;

/* Nonbonded list update */

/* Sort all atoms into small subboxes such that atoms in any box
 * can interact at most with particles in a neighbouring box */
static int
nblist_box(PyNonbondedListObject *nblist, vector3 *x, int natoms,
	   long *subset, int n_sub)
{
  vector3 box1, box2;
  double box_size[3];
  int *p;
  int i, ix, iy, iz, n;

  if (nblist->box_number == NULL) {
    nblist->box_number = (int *)malloc(2*natoms*sizeof(int));
    if (nblist->box_number == NULL) {
      PyErr_NoMemory();
      return 0;
    }
    nblist->box_atoms = nblist->box_number + natoms;
  }
  
  nblist->universe_spec->bounding_box_function(&box1, &box2, x, natoms,
				      nblist->universe_spec->geometry_data);
#if 0
  printf("box1: %lf, %lf, %lf\n", box1[0], box1[1], box1[2]);
  printf("box2: %lf, %lf, %lf\n", box2[0], box2[1], box2[2]);
  printf("cutoff: %lf\n", nblist->cutoff);
#endif
  if (nblist->cutoff > 0.) {
    nblist->box_count[0] = (int)((box2[0]-box1[0])/nblist->cutoff);
    nblist->box_count[1] = (int)((box2[1]-box1[1])/nblist->cutoff);
    nblist->box_count[2] = (int)((box2[2]-box1[2])/nblist->cutoff);
    if (nblist->box_count[0] == 0) nblist->box_count[0] = 1;
    if (nblist->box_count[1] == 0) nblist->box_count[1] = 1;
    if (nblist->box_count[2] == 0) nblist->box_count[2] = 1;
  }
  else
    nblist->box_count[0] = nblist->box_count[1] = nblist->box_count[2] = 1;
  box_size[0] = (box2[0]-box1[0])/nblist->box_count[0];
  box_size[1] = (box2[1]-box1[1])/nblist->box_count[1];
  box_size[2] = (box2[2]-box1[2])/nblist->box_count[2];
#if 0
  printf("division: %d/%d/%d\n", nblist->box_count[0],
	 nblist->box_count[1], nblist->box_count[2]);
  printf("cell size: %lf, %lf, %lf\n", box_size[0], box_size[1], box_size[2]);
#endif

  if (nblist->box_count[0]*nblist->box_count[1]*nblist->box_count[2] >
        nblist->nboxes) {
    free(nblist->boxes);
    nblist->nboxes = nblist->box_count[0] * nblist->box_count[1] *
                     nblist->box_count[2];
    nblist->boxes = malloc(nblist->nboxes*sizeof(nbbox));
    if (nblist->boxes == NULL) {
      PyErr_NoMemory();
      return 0;
    }
  }

  i = 0;
  for (iz = 0; iz < nblist->box_count[2]; iz++)
    for (iy = 0; iy < nblist->box_count[1]; iy++)
      for (ix = 0; ix < nblist->box_count[0]; ix++) {
	nblist->boxes[i].ix = ix;
	nblist->boxes[i].iy = iy;
	nblist->boxes[i].iz = iz;
	nblist->boxes[i].n = 0;
	nblist->boxes[i].i = 0;
	i++;
      }

  n = (n_sub == 0) ? natoms : n_sub;
  for (i = 0; i < n; i++) {
    int box = 0;
    int ai, n;
    ai = (n_sub == 0) ? i : subset[i];
    box = (int)((x[ai][0]-box1[0])/box_size[0]);
    if (box == nblist->box_count[0]) box--;
    n = (int)((x[ai][1]-box1[1])/box_size[1]);
    if (n == nblist->box_count[1]) n--;
    box += nblist->box_count[0]*n;
    n = (int)((x[ai][2]-box1[2])/box_size[2]);
    if (n == nblist->box_count[2]) n--;
    box += nblist->box_count[0]*nblist->box_count[1]*n;
    nblist->box_number[ai] = box;
    nblist->boxes[box].n++;
  }
  p = nblist->box_atoms;
  for (i = 0; i < nblist->nboxes; i++) {
    nblist->boxes[i].atoms = p;
    nblist->boxes[i].i = 0;
    p += nblist->boxes[i].n;
  }
  for (i = 0; i < n; i++) {
    int ai = (n_sub == 0) ? i : subset[i];
    nbbox *box = nblist->boxes + nblist->box_number[ai];
    box->atoms[box->i++] = ai;
  }  
  for (i = 0; i < nblist->nboxes; i++)
    nblist->boxes[i].i = 0;

#if 1
  ix = 0;
  for (i = 0; i < nblist->nboxes; i++) {
    int j;
    for (j = 0; j < nblist->boxes[i].n; j++) {
      int atom = nblist->boxes[i].atoms[j];
      ix++;
      if (nblist->box_number[atom] != i)
	printf("Box number error: %d: %d - %d\n",
	       atom, i, nblist->box_number[atom]);
    }
  }
  if (ix != natoms)
    printf("Atom number error: %d - %d\n", ix, natoms);
#endif

  return 1;
}

/* Check if a given pair is on the excluded or 1-4 list.
 * Successive calls to this routine *must* present the pairs
 * ordered by first-then-second particle number. */
struct pair_iterator
{
  int i_ex, i_14;
};

enum pair_types {NORMAL, ONE_FOUR, EXCLUDED};

static enum pair_types
check_pair(long *excluded, int n_ex, long *one_four, int n_14,
	   struct pair_iterator *iterator, int i1, int i2)
{
  while (iterator->i_ex < n_ex && excluded[iterator->i_ex] < i1)
    iterator->i_ex += 2;
  while (iterator->i_ex < n_ex && excluded[iterator->i_ex] == i1
	 && excluded[iterator->i_ex+1] < i2)
    iterator->i_ex += 2;
  while (iterator->i_14 < n_14 && one_four[iterator->i_14] < i1)
    iterator->i_14 += 2;
  while (iterator->i_14 < n_14 && one_four[iterator->i_14] == i1
	 && one_four[iterator->i_14+1] < i2)
    iterator->i_14 += 2;
  if (iterator->i_ex < n_ex && excluded[iterator->i_ex] == i1
      && excluded[iterator->i_ex+1] == i2)
    return EXCLUDED;
  if (iterator->i_14 < n_14 && one_four[iterator->i_14] == i1
      && one_four[iterator->i_14+1] == i2)
    return ONE_FOUR;
  return NORMAL;
}

/* Increase the memory space allocated for a nonbonded list */
static int
nblist_grow(PyNonbondedListObject *nblist)
{
  const int min_alloc = 1000;
  atom_pair *new;
  int i, alloc, new_pairs;
  new_pairs = nblist->allocated_pairs/10;
  if (new_pairs < min_alloc)
    new_pairs = min_alloc;
  alloc = nblist->allocated_pairs + new_pairs;
  new = (atom_pair *)realloc(nblist->pairs, alloc*sizeof(atom_pair));
  if (new == NULL)
    return 0;
  nblist->pairs = new;
  nblist->allocated_pairs = alloc;
  return 1;
}

/* Add a pair to the nonbonded list */
static int
add_pair(PyNonbondedListObject *nblist, int i1, int i2, int one_four_flag,
	 vector3 d, double cutoff_sq, double check_cutoff_sq)
{
  double d_sq;
  atom_pair pair;

  d_sq = vector_length_sq(d);
  if (cutoff_sq == 0. || d_sq <= cutoff_sq) {
    if (nblist->npairs == nblist->allocated_pairs && !nblist_grow(nblist)) {
      PyErr_NoMemory();
      return 0;
    }
    pair.i1 = i1; pair.i2 = i2;
    pair.r_sq = d_sq;
    pair.check = d_sq > check_cutoff_sq;
    pair.one_four = one_four_flag;
    nblist->pairs[nblist->npairs] = pair;
    nblist->npairs++;
  }
  return 1;
}

/* Update the nonbonded list and reconstruct it if necessary */
int
nblist_update(PyNonbondedListObject *nblist, int natoms, double *coordinates)
{
  vector3 *x = (vector3 *)coordinates;

  if (nblist->remake) {

    long *excluded = (long *)((PyArrayObject *)nblist->excluded_pairs)->data;
    int n_ex = 2*((PyArrayObject *)nblist->excluded_pairs)->dimensions[0];
    long *one_four = (long *)((PyArrayObject *)nblist->one_four_pairs)->data;
    int n_14 = 2*((PyArrayObject *)nblist->one_four_pairs)->dimensions[0];
    long *subset = (long *)((PyArrayObject *)nblist->atom_subset)->data;
    int n_sub = ((PyArrayObject *)nblist->atom_subset)->dimensions[0];
    double cutoff_sq = sqr(nblist->cutoff*safety_factor);
    double check_cutoff_sq = sqr(nblist->cutoff*check_limit);

#if 1
    struct pair_iterator pi;
    int ox, oy, oz;
    int ibox;

#if 0
    {
      int i, j;
      int n = 0;
      for (i = 0; i < natoms; i++)
	for (j = i+1; j < natoms; j++) {
	  vector3 d;
	  double l;
	  nblist->universe_spec->distance_function(d, x[i], x[j],
				    nblist->universe_spec->geometry_data );
	  l = vector_length_sq(d);
	  if (l <= sqr(nblist->cutoff)) {
	    /*	    printf("%d - %d: %lf\n", i, j, sqrt(l)); */
	    n++;
	  }
	}
      printf("%d pairs within cutoff %lf\n", n, nblist->cutoff);
      printf("- %d excluded pairs leaves %d pairs\n", n_ex/2, n-n_ex/2);
    }
#endif
    if (!nblist_box(nblist, x, natoms, subset, n_sub))
      return 0;
    nblist->npairs = 0;
    for (ibox = 0; ibox < nblist->nboxes; ibox++) {
      nbbox *box = &nblist->boxes[ibox];
      int *atoms = box->atoms;
      int i, j;

      box->i = 1;

      pi.i_ex = 0; pi.i_14 = 0;
      for (i = 0; i < box->n; i++)
	for (j = i+1; j < box->n; j++) {
	  int ai = atoms[i];
	  int aj = atoms[j];
	  enum pair_types type = check_pair(excluded, n_ex, one_four, n_14,
					    &pi, ai, aj);
	  if (type != EXCLUDED) {
	    vector3 d;
	    nblist->universe_spec->distance_function(d, x[ai], x[aj],
				       nblist->universe_spec->geometry_data );
	    if (!add_pair(nblist, ai, aj, type == ONE_FOUR,
			  d, cutoff_sq, check_cutoff_sq))
	      return 0;
	  }
	}

      for (ox = -1; ox < 2; ox++)
	for (oy = -1; oy < 2; oy++)
	  for (oz = -1; oz < 2; oz++)
	    if (!(ox == 0 && oy == 0 && oz == 0)) {
	      int ix = box->ix+ox;
	      int iy = box->iy+oy;
	      int iz = box->iz+oz;
	      int ibox;
	      nbbox *other;
	      if (nblist->universe_spec->is_periodic) {
		if (ix < 0) ix += nblist->box_count[0];
		if (iy < 0) iy += nblist->box_count[1];
		if (iz < 0) iz += nblist->box_count[2];
		if (ix >= nblist->box_count[0]) ix -= nblist->box_count[0];
		if (iy >= nblist->box_count[1]) iy -= nblist->box_count[1];
		if (iz >= nblist->box_count[2]) iz -= nblist->box_count[2];
	      }
	      else if (ix < 0 || iy < 0 || iz < 0
		       || ix >= nblist->box_count[0]
		       || iy >= nblist->box_count[1]
		       || iz >= nblist->box_count[2])
		continue;
	      ibox = ix + nblist->box_count[0]*(iy + nblist->box_count[1]*iz);
#if 1
	      if (ibox < 0 || ibox >= nblist->nboxes)
		printf("Illegal box number: %d\n", ibox);
#endif
	      other = &nblist->boxes[ibox];
#if 1
	      if (ix != other->ix || iy != other->iy || iz != other->iz)
		printf("Wrong box: %d,%d,%d - %d,%d,%d\n",
		       ix, iy, iz, other->ix, other->iy, other->iz);
#endif
	      if (other->i) continue;
	      pi.i_ex = 0; pi.i_14 = 0;
	      i = 0; j = 0;
	      while (i < box->n && j < other->n) {
		if (atoms[i] <= other->atoms[j]) {
		  int k;
		  for (k = j; k < other->n; k++) {
		    int ai = atoms[i];
		    int aj = other->atoms[k];
		    enum pair_types type = check_pair(excluded, n_ex,
						      one_four, n_14,
						      &pi, ai, aj);
		    if (type != EXCLUDED) {
		      vector3 d;
		      nblist->universe_spec->distance_function(d, x[ai], x[aj],
				       nblist->universe_spec->geometry_data);
		      if (!add_pair(nblist, ai, aj, type == ONE_FOUR,
				    d, cutoff_sq, check_cutoff_sq))
			return 0;
		    }
		  }
		  i++;
		}
		else {
		  int k;
		  for (k = i; k < box->n; k++) {
		    int ai = other->atoms[j];
		    int aj = atoms[k];
		    enum pair_types type = check_pair(excluded, n_ex,
						      one_four, n_14,
						      &pi, ai, aj);
		    if (type != EXCLUDED) {
		      vector3 d;
		      nblist->universe_spec->distance_function(d, x[ai], x[aj],
				       nblist->universe_spec->geometry_data);
		      if (!add_pair(nblist, ai, aj, type == ONE_FOUR,
				    d, cutoff_sq, check_cutoff_sq))
			return 0;
		    }
		  }
		  j++;
		}
	      }
	    }
    }
#if 0
    {
      int i;
      atom_pair *pair = nblist->pairs;
      int n = 0;
      for (i = 0; i < nblist->npairs; i++, pair++)
	if (pair->r_sq <= sqr(nblist->cutoff))
	  n++;
      printf("%d pairs within cutoff in nonbonded list\n", n);
    }
#endif
#else
    double *distance_data = nblist->universe_spec->geometry_data;
    vector3 d;
    int i_ex = 0, i_14 = 0;
    int n, i1, i2, j1, j2;
    atom_pair pair;
    double d_sq;

    nblist->npairs = 0;
    n = (n_sub == 0) ? natoms : n_sub;
    if (nblist->universe_spec->distance_function == distance_vector_pointer) {
#     define distance_vector_function distance_vector_1
#     include "nonbonded1.i"
#     undef distance_vector_function
    }
    else if (nblist->universe_spec->distance_function ==
	                             orthorhombic_distance_vector_pointer) {
#     define distance_vector_function distance_vector_2
#     include "nonbonded1.i"
#     undef distance_vector_function
    }
    else {
#     define distance_vector_function nblist->universe_spec->distance_function
#     include "nonbonded1.i"
#     undef distance_vector_function
    }
#endif
    nblist->remake = 0;
    nblist->countdown = remake_frequency;
  }
  else {

    double *distance_data = nblist->universe_spec->geometry_data;
    atom_pair *pair = nblist->pairs;
    double real_cutoff_sq = sqr(nblist->cutoff);
    vector3 d;
    int i;

    if (nblist->universe_spec->distance_function == distance_vector_pointer) {
#     define distance_vector_function distance_vector_1
#     include "nonbonded2.i"
#     undef distance_vector_function
    }
    else if (nblist->universe_spec->distance_function ==
	                             orthorhombic_distance_vector_pointer) {
#     define distance_vector_function distance_vector_2
#     include "nonbonded2.i"
#     undef distance_vector_function
    }
    else {
#     define distance_vector_function nblist->universe_spec->distance_function
#     include "nonbonded2.i"
#     undef distance_vector_function
    }
    if (--nblist->countdown == 0)
      nblist->remake = 1;
  }
  return 1;
}

/* Pseudo forcefield that updates a non-bonded list */

double
nonbonded_evaluator(PyObject *object, int natoms,
		    PyArrayObject *coordinates,
		    PyObject *gradients, PyObject *force_constants,
		    gradient_function *gradient_fn, fc_function *fc_fn,
		    int small_change)
{
  PyFFEvaluatorObject *self = (PyFFEvaluatorObject *)object;
  PyNonbondedListObject *nblist = (PyNonbondedListObject *)self->data1;
  vector3 *x = (vector3 *)coordinates->data;

  if (!small_change)
    nblist->remake = 1;
  nblist_update(nblist, natoms, (double *)x);
  self->energy = 0.;
  self->virial = 0.;
  self->virial_available = 1;
  return 0.;
}

/* Pair contribution to nonbonded interactions; written as a macro
   for efficiency */

#define pair_lj_es_energy(eps, sigma, qiqj, distance_vector_function) \
{ \
  vector3 rij; \
  double r_sq; \
  double deriv = 0.; \
  double deriv2 = 0.; \
  distance_vector_function(rij, x[j], x[i], distance_data); \
  r_sq = vector_length_sq(rij); \
  if ((eps) != 0.) { \
    double sr2 = sqr((sigma))/r_sq; \
    double sr6 = cube(sr2); \
    double sr12 = sqr(sr6); \
    double d = -24.*(eps)*(2.*sr12-sr6)/r_sq; \
    self->out1 += 4.*(eps)*(sr12-sr6); \
    self->out3 -= d*r_sq; \
    deriv += d; \
    deriv2 += 24.*(eps)*(26.*sr12-7.*sr6)/r_sq; \
  } \
  if ((qiqj) != 0.) { \
    double r = sqrt(r_sq); \
    double d = -(qiqj)/(r*r_sq); \
    self->out2 += (qiqj)/r; \
    self->out4 += (qiqj)/r; \
    deriv += d; \
    deriv2 += 2.*(qiqj)/(r*r_sq); \
  } \
  if (gradients != NULL) { \
    vector3 grad; \
    grad[0] = deriv*rij[0]; \
    grad[1] = deriv*rij[1]; \
    grad[2] = deriv*rij[2]; \
    if (gradient_fn != NULL) { \
      (*gradient_fn)(self, gradients, i, grad); \
      vector_changesign(grad); \
      (*gradient_fn)(self, gradients, j, grad); \
    } \
    else { \
      vector3 *f = (vector3 *)((PyArrayObject *)gradients)->data; \
      f[i][0] += grad[0]; \
      f[i][1] += grad[1]; \
      f[i][2] += grad[2]; \
      f[j][0] -= grad[0]; \
      f[j][1] -= grad[1]; \
      f[j][2] -= grad[2]; \
    } \
  } \
  if (force_constants != NULL) \
    add_pair_fc(self, force_constants, fc_fn, \
		i, j, rij, r_sq, deriv, deriv2); \
}


/* Combined Lennard-Jones and electrostatic interactions,
   used only for infinite cutoff */

double
lj_es_evaluator(PyObject *object, int natoms,
		PyArrayObject *coordinates,
		PyObject *gradients, PyObject *force_constants,
		gradient_function *gradient_fn, fc_function *fc_fn,
		int small_change)
{
  PyFFEvaluatorObject *self = (PyFFEvaluatorObject *)object;
  vector3 *x = (vector3 *)coordinates->data;
  distance_fn *d_fn = self->universe_spec->distance_function;
  double *distance_data = self->universe_spec->geometry_data;

  PyObject *pair_lists = self->data1;
  PyArrayObject *array_ex = (PyArrayObject *)PyList_GetItem(pair_lists, 0);
  PyArrayObject *array_14 = (PyArrayObject *)PyList_GetItem(pair_lists, 1);
  PyArrayObject *array_subset =
                  (PyArrayObject *)PyList_GetItem(pair_lists, 2);
  long *excluded = (long *)array_ex->data;
  long n_ex = 2*array_ex->dimensions[0];
  long *one_four = (long *)array_14->data;
  long n_14 = 2*array_14->dimensions[0];
  long *subset = (long *)array_subset->data;
  long n_sub = array_subset->dimensions[0];
  double *eps_sigma = (double *)((PyArrayObject *)self->data2)->data;
  int ntypes = ((PyArrayObject *)self->data2)->dimensions[0];
  long *lj_type = (long *)((PyArrayObject *)self->data3)->data;
  double *charge = (double *)((PyArrayObject *)self->data4)->data;
  double lj_one_four_factor = self->param1;
  double es_one_four_factor = self->param2;
  int k, l, n;
  
  self->energy = 0.;
  self->virial = 0.;
  self->virial_available = 1;
  self->out1 = self->out2 = self->out3 = self->out4 = 0.;
  if (d_fn == distance_vector_pointer) {
#   define distance_vector_function distance_vector_1
#   include "nonbonded3.i"
#   undef distance_vector_function
  }
  else if (d_fn == orthorhombic_distance_vector_pointer) {
#   define distance_vector_function distance_vector_2
#   include "nonbonded3.i"
#   undef distance_vector_function
  }
  else {
#   define distance_vector_function (*d_fn)
#   include "nonbonded3.i"
#   undef distance_vector_function
  }
  self->energy = self->out1 + self->out2;
  self->virial = self->out3 + self->out4;
  return self->energy;
}


/* Lennard-Jones interactions */

double
lennard_jones_evaluator(PyObject *object, int natoms,
			PyArrayObject *coordinates,
			PyObject *gradients, PyObject *force_constants,
			gradient_function *gradient_fn, fc_function *fc_fn,
			int small_change)
{
  PyFFEvaluatorObject *self = (PyFFEvaluatorObject *)object;
  vector3 *x = (vector3 *)coordinates->data;
  distance_fn *d_fn = self->universe_spec->distance_function;
  double *distance_data = self->universe_spec->geometry_data;

  PyNonbondedListObject *nblist = (PyNonbondedListObject *)self->data1;
  double *eps_sigma = (double *)((PyArrayObject *)self->data2)->data;
  int ntypes = ((PyArrayObject *)self->data2)->dimensions[0];
  long *lj_type = (long *)((PyArrayObject *)self->data3)->data;
  double cutoff_sq = sqr(self->param1);
  double one_four_factor = self->param2;
  double energy = 0.;
  atom_pair *pair = nblist->pairs;
  int k;

  self->energy = 0.;
  self->virial = 0.;
  self->virial_available = 1;
  if (d_fn == distance_vector_pointer) {
#   define distance_vector_function distance_vector_1
#   include "nonbonded5.i"
#   undef distance_vector_function
  }
  else if (d_fn == orthorhombic_distance_vector_pointer) {
#   define distance_vector_function distance_vector_2
#   include "nonbonded5.i"
#   undef distance_vector_function
  }
  else {
#   define distance_vector_function (*d_fn)
#   include "nonbonded5.i"
#   undef distance_vector_function
  }
  return self->energy;
}


/* Electrostatic interactions */

double
electrostatic_evaluator(PyObject *object, int natoms,
			PyArrayObject *coordinates,
			PyObject *gradients, PyObject *force_constants,
			gradient_function *gradient_fn, fc_function *fc_fn,
			int small_change)
{
  PyFFEvaluatorObject *self = (PyFFEvaluatorObject *)object;
  vector3 *x = (vector3 *)coordinates->data;
  distance_fn *d_fn = self->universe_spec->distance_function;
  double *distance_data = self->universe_spec->geometry_data;

  PyNonbondedListObject *nblist = (PyNonbondedListObject *)self->data1;
  double *charge = (double *)((PyArrayObject *)self->data2)->data;
  double cutoff_sq = sqr(self->param1);
  double one_four_factor = self->param2;
  atom_pair *pair = nblist->pairs;
  int k;

  self->energy = 0.;
  self->virial = 0.;
  self->virial_available = 1;
  if (d_fn == distance_vector_pointer) {
#   define distance_vector_function distance_vector_1
#   include "nonbonded6.i"
#   undef distance_vector_function
  }
  else if (d_fn == orthorhombic_distance_vector_pointer) {
#   define distance_vector_function distance_vector_2
#   include "nonbonded6.i"
#   undef distance_vector_function
  }
  else {
#   define distance_vector_function (*d_fn)
#   include "nonbonded6.i"
#   undef distance_vector_function
  }
  return self->energy;
}

#ifdef WITH_DPMTA
/* Electrostatic interactions via multipoles */
double
es_mp_evaluator(PyObject *object, int natoms,
		PyArrayObject *coordinates,
		PyObject *gradients, PyObject *force_constants,
		gradient_function *gradient_fn, fc_function *fc_fn,
		int small_change)
{
  PyFFEvaluatorObject *self = (PyFFEvaluatorObject *)object;
  vector3 *x = (vector3 *)coordinates->data;
  distance_fn *d_fn = self->universe_spec->distance_function;
  double *distance_data = self->universe_spec->geometry_data;
  PmtaInitData *initdata;
  static PmtaInitData *last_initdata = NULL;
  PmtaParticle *particle;
  PmtaPartInfo *force;
  PmtaVector virial;
  double potential;
  PyObject *pair_lists = self->data1;
  PyArrayObject *array_ex = (PyArrayObject *)PyList_GetItem(pair_lists, 0);
  PyArrayObject *array_14 = (PyArrayObject *)PyList_GetItem(pair_lists, 1);
  PyArrayObject *array_subset =
                  (PyArrayObject *)PyList_GetItem(pair_lists, 2);
  long *excluded = (long *)array_ex->data;
  long n_ex = 2*array_ex->dimensions[0];
  long *one_four = (long *)array_14->data;
  long n_14 = 2*array_14->dimensions[0];
  long *subset = (long *)array_subset->data;
  long n_sub = array_subset->dimensions[0];
  double *charge = (double *)((PyArrayObject *)self->data2)->data;
  double es_one_four_factor = self->param1;
  int k, l, n;

  n = (n_sub == 0) ? natoms : n_sub;
  initdata = (PmtaInitData *)self->scratch;
  particle = (PmtaParticle *)(initdata+1);
  force = (PmtaPartInfo *)(particle+n);
  if (initdata != last_initdata) {
    PMTAinit(initdata, NULL);
    last_initdata = initdata;
  }
  for (k = 0; k < n; k++) {
    int i = (n_sub == 0) ? k : subset[k];
    particle[k].p.x = x[i][0];
    particle[k].p.y = x[i][1];
    particle[k].p.z = x[i][2];
    particle[k].q = charge[i];
  }
  if (self->param2 == 0.) {
    PmtaVector v1, v2, v3, center;
    double xmin, xmax, ymin, ymax, zmin, zmax;
    xmin = ymin = zmin = undefined;
    xmax = ymax = zmax = -undefined;
    for (k = 0; k < n; k++) {
      int i = (n_sub == 0) ? k : subset[k];
      if (x[i][0] < xmin) xmin = x[i][0];
      if (x[i][1] < ymin) ymin = x[i][1];
      if (x[i][2] < zmin) zmin = x[i][2];
      if (x[i][0] > xmax) xmax = x[i][0];
      if (x[i][1] > ymax) ymax = x[i][1];
      if (x[i][2] > zmax) zmax = x[i][2];
    }
    v1.x = xmax-xmin; v1.y = 0.; v1.z = 0.;
    v2.x = 0.; v2.y = ymax-ymin; v2.z = 0.;
    v3.x = 0.; v3.y = 0.; v3.z = zmax-zmin;
    center.x = 0.5*(xmin+xmax);
    center.y = 0.5*(ymin+ymax);
    center.z = 0.5*(zmin+zmax);
    PMTAresize(&v1, &v2, &v3, &center);
  }
  else {
    int change = 0;
    if (distance_data[0] != initdata->v1.x) {
      initdata->v1.x = distance_data[0];
      change = 1;
    }
    if (distance_data[1] != initdata->v2.y) {
      initdata->v2.y = distance_data[1];
      change = 1;
    }
    if (distance_data[2] != initdata->v3.z) {
      initdata->v3.z = distance_data[2];
      change = 1;
    }
    if (change)
      PMTAresize(&initdata->v1, &initdata->v2, &initdata->v3,
		 &initdata->cellctr);
  }
  PMTAforce(n, particle, force, NULL);
  PMTAvirial(&potential, &virial, NULL, NULL);
  for (k = 0; k < n; k++) {
    int i = (n_sub == 0) ? k : subset[k];
    vector3 grad;
    grad[0] = -electrostatic_energy_factor*force[k].f.x;
    grad[1] = -electrostatic_energy_factor*force[k].f.y;
    grad[2] = -electrostatic_energy_factor*force[k].f.z;
    if (gradients != NULL) {
      if (gradient_fn != NULL)
	(*gradient_fn)(self, gradients, i, grad);
      else {
	vector3 *f = (vector3 *)((PyArrayObject *)gradients)->data;
	f[i][0] += grad[0];
	f[i][1] += grad[1];
	f[i][2] += grad[2];
      }
    }
  }
  self->virial = -electrostatic_energy_factor*(virial.x+virial.y+virial.z);
  self->energy = potential*electrostatic_energy_factor;
  if (force_constants != NULL)
    self->error = 1;
  self->out1 = self->out2 = self->out3 = self->out4 = 0.;
  if (d_fn == distance_vector_pointer) {
#   define distance_vector_function distance_vector_1
#   include "nonbonded4.i"
#   undef distance_vector_function
  }
  else if (d_fn == orthorhombic_distance_vector_pointer) {
#   define distance_vector_function distance_vector_2
#   include "nonbonded4.i"
#   undef distance_vector_function
  }
  else {
#   define distance_vector_function (*d_fn)
#   include "nonbonded4.i"
#   undef distance_vector_function
  }
  self->virial_available = 1;
  self->virial += self->out4;
  self->energy += self->out2;
  return self->energy;
}
#endif
