# This module implements surface and volume calculations based on
# Frank Eisenhaber's NSC code.
#
# Written by Lutz Ehrlich
# Adapted to MMTK conventions by Konrad Hinsen
# last revision: 1999-3-16
#

import Collection, Units
import Numeric
import MMTK_surface

#
# Surface and volume calculation for an arbitrary group of atoms.
#
def surfaceAndVolume(object, probe_radius = 0.):
    atoms = object.atomList()
    natoms = len(atoms)
    coordinates = Numeric.zeros((natoms, 3), Numeric.Float)
    radii = Numeric.zeros((natoms,), Numeric.Float)
    for i in range(natoms):
	a = atoms[i]
	coordinates[i] = a.position().array
	radii[i] = a.vdW_radius
    coordinates.shape = (3*natoms,)
    [surface, volume] = \
	      MMTK_surface.getSurfaceAndVolume(coordinates, radii+probe_radius)
    return surface, volume


#
# Determination of surface atoms and exposed surface per atom
#
def surfaceAtoms(object, probe_radius = 0.):
    atoms = object.atomList()
    natoms = len(atoms)
    coordinates = Numeric.zeros((natoms, 3), Numeric.Float)
    radii = Numeric.zeros((natoms,), Numeric.Float)
    for i in range(natoms):
	a = atoms[i]
	coordinates[i] = a.position().array
	radii[i] = a.vdW_radius
    coordinates.shape = (3*natoms,)

    area_list = MMTK_surface.getSurfaceAtoms(coordinates, radii+probe_radius)
    if len(area_list) != natoms:
	raise ValueError, 'number of atoms not equal number of areas'

    surface_atoms = {}
    for i in range(natoms):
	if area_list[i] > 0.:
	    # we have a surface atom
	    surface_atoms[atoms[i]] = area_list[i]

    return surface_atoms

#
# Determination of contacts between two groups of atoms
#
def findContacts(object1, object2, contact_factor = 1., cutoff= 0.):
    atoms1 = object2.atomList()
    atoms2 = object2.atomList()
    index1 = map( lambda x: x.index, atoms1)
    radius1 = map( lambda x, c=cutoff: x.vdW_radius+0.5*c, atoms1)
    index2 = map( lambda x: x.index, atoms2)
    radius2 = map( lambda x, c=cutoff: x.vdW_radius+0.5*c, atoms2)
    contacts = MMTK_surface.getContacts(self.universe().configuration().array,
                                        contact_factor, radius1, index1,
                                        radius2, index2)
    contact_pairs = []
    for [i, j] in contacts: 
	contact_pairs.append(Contact(atoms1[i], atoms2[j]))
    return contact_pairs

#
# A class representing a contact (two atoms plus their distance).
#
class Contact:
    
    def __init__(self, atom1, atom2):
	if atom1 is atom2:
	    raise ValueError, 'no contacts between equal atoms'
	self.a1 = atom1
	self.a2 = atom2

    def __cmp__(self, other):
	"""
	two contacts are the same if they comprise the same atom objects. 
	A contact is 'greater' than another contact if the atoms are further
	apart.
	"""
	try:
	    self.dist
	except AttributeError:
	    self.dist = (atom1.position() - atom2.position()).length()

	if ((self.a1 is other.a1) and (self.a2 is other.a2)) or \
	   ((self.a1 is other.a2) and (self.a2 is other.a1)):
	    return 0
	elif self.dist < other.dist:
	    return -1
	elif self.dist > other.dist:
	    return 1
	else:
	    return cmp(id(self.a1), id(self.a2))

    def __getitem__(self, item):
	return (self.a1, self.a2)[item]

    def __hash__(self):
	return id(self.a1) + id(self.a2)

    def __str__(self):
	return str((self.a1, self.a2))

    def __repr__(self):
	return '<Contact' + str(self) + '>'
